/*
 * Copyright © 2019-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/statvfs.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <librzrsdk2/redzone_c_api.h>
#include <librzrsdk2/props_api.h>

#include "enatsc3source.h"
#include "airwavzrzrsdk2source.h"

static char *ensys = "bb2fmp4";
static char *ensubsys = "bbrzrsdk2src";

#define MAX_BASEBAND_PAYLOAD_SIZE 65536

#define MAX_BASEBAND_FILE_SOURCES 32

typedef struct
{
    struct _ENATSC3Source base;

    RedZoneCaptureHandle_t hRedZoneCapture;

    int32_t fOpened;

    enum RedZoneOperatingMode operatingMode;
    int32_t frequencyKHz;

    // The currently tuned PLPs and their slot assignemnts. Most ATSC 3 demods
    // can handle only 4 PLPs, as that is all that is required by spec.
    uint8_t vPLPIds[4];
    uint64_t plpIdMask;

    int32_t fPLPIdMaskChanged;
    ENATSC3ALPHandle_t hATSC3ALP;

    uint64_t basebandPacketNumber;
    struct timespec basebandPacketTime;
    uint8_t *pBasebandPayload;

    ENATSC3SourcePLPsCallbackFunction PLPsCallback;
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    ENATSC3SourceBasebandPacketCallbackFunction BasebandPacketCallback;
    void *pCallbackUserData;

    struct _stats
    {
        uint8_t masterLock;
        uint64_t plpValidMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        // Total packets received
        uint32_t nBasebandPackets;
        // Total bits received
        uint32_t nBytes;
        // Number of packets containing uncorrected errors.
        uint32_t nErroredPackets;

        struct _plpStats
        {
            // PLP packets received
            uint32_t nBasebandPackets;
            // PLP bits received
            uint32_t nBytes;
            // Number of continuity counter sequence discontinuities detected
            uint32_t nContinuityCounterErrors;
            // Number of packets containing uncorrected errors.
            uint32_t nErroredPackets;
            int32_t lastContinuityCounter;
        } vPLPStats[64];
    } stats;

    pthread_mutex_t statsLock;
    pthread_mutex_t lock;
} ENATSC3BasebandRZRSDK2Source;

struct
{
    int32_t fInitialized;

    ENATSC3BasebandRZRSDK2Source vSources[MAX_BASEBAND_FILE_SOURCES];
    int32_t nSources;

    pthread_mutex_t lock;
} atsc3BasebandRZRSDK2SourceGlobalsG =
{
    .fInitialized = 0,
};

void
ATSC3BasebandRZRSDK2SourceBasebandPacketCallback(
    RedZoneCaptureBasebandPacket *pPacket, void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)pUserData;
    struct _plpStats *pPLPStats;
    int32_t fCCError = 0;
    int32_t nPayloadBytes = 0;
    struct timespec packetTime;
    int32_t i;

    if (!pSource->fOpened)
    {
        ERRORPRINT("Baseband packet on closed source?\n ");
        goto out;
    }

    if (pPacket->v0.plpId > 64)
    {
        ERRORPRINT("Invalid PLP Id %d\n", pPacket->v0.plpId);
        pSource->stats.nBasebandPackets++;
        pSource->stats.nErroredPackets++;
        for (i = 0; i < pPacket->v0.nVecs; i++)
        {
            pSource->stats.nBytes += pPacket->v0.vVecs[i].length;
        }
        goto out;
    }

    clock_gettime(CLOCK_REALTIME, &packetTime);

    pPLPStats = pSource->stats.vPLPStats + pPacket->v0.plpId;
    pPLPStats->nBasebandPackets++;
    pSource->stats.nBasebandPackets++;
    for (i = 0; i < pPacket->v0.nVecs; i++)
    {
        pPLPStats->nBytes += pPacket->v0.vVecs[i].length;
        pSource->stats.nBytes += pPacket->v0.vVecs[i].length;
    }
    if (pPLPStats->lastContinuityCounter != -1 &&
        ((pPLPStats->lastContinuityCounter + 1) & 0xff) !=
        pPacket->v0.continuityCounter)
    {
        pPLPStats->nContinuityCounterErrors++;
        fCCError = 1;
    }
    pPLPStats->lastContinuityCounter = pPacket->v0.continuityCounter;
    if (pPacket->v0.fError)
    {
        pPLPStats->nErroredPackets++;
        pSource->stats.nErroredPackets++;
    }

    if (pSource->plpIdMask & (1ULL << pPacket->v0.plpId))
    {
        ENATCS3L1Time l1time;

        for (i = 0; i < pPacket->v0.nVecs; i++)
        {
            if (pPacket->v0.vVecs[i].length + nPayloadBytes > 
                MAX_BASEBAND_PAYLOAD_SIZE)
            {
                ERRORPRINT("Baseband packet length overflow %d\n",
                    pPacket->v0.vVecs[i].length + nPayloadBytes);
                break;
            }

            memcpy(pSource->pBasebandPayload + nPayloadBytes, 
                pPacket->v0.vVecs[i].pData, pPacket->v0.vVecs[i].length);
            nPayloadBytes += pPacket->v0.vVecs[i].length;
        }

        l1time.precision = pPacket->v0.L1TimePrecision;
        l1time.seconds = pPacket->v0.L1TimeSeconds;
        l1time.milliseconds = pPacket->v0.L1TimeMilliseconds;
        l1time.microseconds = pPacket->v0.L1TimeMicroseconds;
        l1time.nanoseconds = pPacket->v0.L1TimeNanoseconds;

        if (pSource->BasebandPacketCallback)
        {
            pSource->BasebandPacketCallback(pPacket->v0.plpId, &packetTime, 
                pSource->basebandPacketNumber, pPacket->v0.fError, fCCError, 
                &l1time, pSource->pBasebandPayload, nPayloadBytes, 
                pSource->pCallbackUserData);
        }
    }

    pSource->basebandPacketNumber++;
    pSource->basebandPacketTime = packetTime;

out:
    return;
}

static inline void
GetStatistics(ENATSC3BasebandRZRSDK2Source *pSource)
{
    struct _stats stats;
    int32_t fHaveMasterLock = 1;
    int32_t fHavePLPValidMask = 1;
    int32_t fHaveLLSValidMask = 1;
    int32_t fHaveSNR = 1;
    int32_t fHaveRSSI = 1;
    uint64_t plpStatsMask = 0;

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneMasterLockProp, &stats.masterLock,
        sizeof(&stats.masterLock) != RZR_SUCCESS))
    {
        fHaveMasterLock = 0;
        ERRORPRINT("Failed to get master lock\n");
    }

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZonePLPValidBitmaskProp, &stats.plpValidMask, 
        sizeof(stats.plpValidMask)) != RZR_SUCCESS)
    {
        fHavePLPValidMask = 0;
        ERRORPRINT("Failed to get plp valid mask\n");
    }

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneLLSValidBitmaskProp, &stats.llsValidMask, 
        sizeof(stats.llsValidMask)) != RZR_SUCCESS)
    {
        fHaveLLSValidMask = 0;
        ERRORPRINT("Failed to get lls plps mask\n");
    }

    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneSNRProp, &stats.snr,
        sizeof(stats.snr)) != RZR_SUCCESS)
    {
        fHaveSNR = 0;
        ERRORPRINT("Failed to get snr\n");
    }
    if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
        RedZoneRSSIProp, &stats.rssi,
        sizeof(stats.rssi)) != RZR_SUCCESS)
    {
        fHaveSNR = 0;
        ERRORPRINT("Failed to get rssi\n");
    }

    struct RedZonePLPSpecificInfo3 vPLPInfo[64];
    uint64_t mask;
    int32_t i;

    for (i = 0, mask = 1; i < 64; i++, mask <<= 1)
    {
        if (stats.plpValidMask & mask)
        {
            vPLPInfo[i].index = 255;
            vPLPInfo[i].plp_id = i;
            if (RedZoneCaptureGetProp(pSource->hRedZoneCapture, 
                RedZonePLPSpecificInfoProp, vPLPInfo + i, sizeof(vPLPInfo[0])))
            {
                ERRORPRINT("Failed to get plp %d info\n", i);
                continue;
            }
        }

        plpStatsMask |= mask;
    }

    RedZoneCaptureStatistics statistics;
    RedZoneCaptureGetStatistics(pSource->hRedZoneCapture, &statistics);

    LOGPRINT_NOFILE("snr: %d rssi: %d Baseband frames: %"PRId64" RF errors: %"PRId64" Continuity errors: %"PRId64" Framing errors:%"PRId64"\n",
        stats.snr, stats.rssi, statistics.nPackets, 
        statistics.nUncorrectedRFErrors,
        statistics.nContinuityCounterErrors, statistics.nFramingErrors);

    if (pthread_mutex_lock(&pSource->statsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (fHaveMasterLock)
    {
       pSource->stats.masterLock = stats.masterLock;
    }

    if (fHavePLPValidMask)
    {
        pSource->stats.plpValidMask = stats.plpValidMask;
    }

    if (fHaveLLSValidMask)
    {
        if (pSource->stats.llsValidMask != stats.llsValidMask)
        {
            pSource->LLSPLPsCallback(stats.llsValidMask, 
                &pSource->basebandPacketTime, 
                pSource->basebandPacketNumber, pSource->pCallbackUserData);
        }
        pSource->stats.llsValidMask = stats.llsValidMask;
    }

    if (fHaveSNR)
    {
        pSource->stats.snr = stats.snr;
    }

    if (fHaveRSSI)
    {
        pSource->stats.rssi = stats.rssi;
    }

    if (pthread_mutex_unlock(&pSource->statsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }
}

ENStatus
ATSC3BasebandRZRSDK2SourceStart(ENATSC3Source hSource, int32_t *pFRunning)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    int32_t i;
    int32_t j;
    struct timespec tNextStats = {.tv_sec = 0, .tv_nsec = 0};
    ENStatus rval = ENStatus_Success;

    memset(&pSource->stats, 0, sizeof(pSource->stats));
    for (i = 0; i < numof(pSource->stats.vPLPStats); i++)
    {
        pSource->stats.vPLPStats[i].lastContinuityCounter = -1;
    }

    if (RedZoneCaptureStart(pSource->hRedZoneCapture))
    {
        ERRORPRINT("Failed to start RedZoneCapture\n");
        rval = ENStatus_Failure;
        goto out;
    }

    while (*pFRunning)
    {
        struct timespec mononow;
        usleep(100000);

        clock_gettime(CLOCK_MONOTONIC, &mononow);
        if (mononow.tv_sec > tNextStats.tv_sec ||
            (mononow.tv_sec == tNextStats.tv_sec &&
             mononow.tv_nsec >= tNextStats.tv_nsec))
        {
            GetStatistics(pSource);
            tNextStats = mononow;
            tNextStats.tv_sec += 1;
        }

        if (pSource->fPLPIdMaskChanged)
        {
            uint64_t plpIdMask;
            uint64_t plpIdMaskTemp;

            // We're going to snapshot the mask since processing the change 
            // could take a while and we don't want to hold the lock that
            // long. If someone changes the mask again we'll catch it the next
            // time around.
            if (pthread_mutex_lock(&pSource->lock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            plpIdMask = pSource->plpIdMask;
            pSource->fPLPIdMaskChanged = 0;

            if (pthread_mutex_unlock(&pSource->lock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            // First reconcile the current PLPs with the mask.
            plpIdMaskTemp = plpIdMask;
            for (i = 0; i < numof(pSource->vPLPIds); i++)
            {
                if (pSource->vPLPIds[i] != 255)
                {
                    if (plpIdMaskTemp & 1ULL << pSource->vPLPIds[i])
                    {
                        // Already set, remove from mask
                        plpIdMaskTemp &= ~1ULL << pSource->vPLPIds[i];
                    }
                    else
                    {
                        // No longer set, free up slot
                        pSource->vPLPIds[i] = 255;
                    }
                }
            }

            // Next assign new PLPs to slots
            for (i = 0; i < 64; i++)
            {
                if (plpIdMaskTemp & 1)
                {
                    for (j = 0; j < numof(pSource->vPLPIds); j++)
                    {
                        if (pSource->vPLPIds[j] == 255)
                        {
                            pSource->vPLPIds[j] = i;
                            break;
                        }
                    }

                    if (j == numof(pSource->vPLPIds))
                    {
                        ERRORPRINT("Too many plps: couldn't set %d, mask %x, slots %d %d %d %d\n", 
                            i, plpIdMask, pSource->vPLPIds[0], 
                            pSource->vPLPIds[1], pSource->vPLPIds[2], 
                            pSource->vPLPIds[3]);
                    }
                }

                plpIdMaskTemp >>=1;
            }

            LOGPRINT_NOCONSOLE("Setting PLP mask 0x%x slots %d %d %d %d\n",
                plpIdMask, pSource->vPLPIds[0], pSource->vPLPIds[1], 
                pSource->vPLPIds[2], pSource->vPLPIds[3]); 

            RedZonePLPSet val;

            val.plp0_id = pSource->vPLPIds[0];
            val.plp1_id = pSource->vPLPIds[1];
            val.plp2_id = pSource->vPLPIds[2];
            val.plp3_id = pSource->vPLPIds[3];

            if (RedZoneCaptureSetProp(pSource->hRedZoneCapture,
                RedZonePLPSelectionProp, &val, sizeof(val)))
            {
                ERRORPRINT("Failed to set PLP selection\n");
            }

            if (pSource->PLPsCallback)
            {
                pSource->PLPsCallback(plpIdMask, pSource->pCallbackUserData);
            }
        }
    }

    if (RedZoneCaptureStop(pSource->hRedZoneCapture))
    {
        ERRORPRINT("Failed to stop RedZoneCapture\n");
        rval = ENStatus_Failure;
        goto out;
    }

out:
    return rval;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceOpen(int32_t deviceNumber, 
    ENATSC3BasebandRZRSDK2SourceOperatingMode operatingMode, 
    int32_t frequencyHz, ENATSC3Source *ppSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources));
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources + deviceNumber;

    if (pSource->fOpened)
    {
        ERRORPRINT("Device %d already opened\n", deviceNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->pBasebandPayload = NULL;
    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // Set all PLPs invalid. Note that we will actually tune the first slot
    // to PLP 0, since one PLP must be set when initially tuning and 0 seems
    // to always be present in emissions.
    pSource->vPLPIds[0] = 255;
    pSource->vPLPIds[1] = 255;
    pSource->vPLPIds[2] = 255;
    pSource->vPLPIds[3] = 255;

    if (operatingMode == ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC)
    {
        pSource->operatingMode = RedZoneOperatingModeATSC;
    }
    else if (operatingMode == ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC3)
    {
        pSource->operatingMode = RedZoneOperatingModeATSC3;
    }
    else
    {
        ERRORPRINT("Invalid operating mode %d\n", operatingMode);
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneOperatingModeProp, &pSource->operatingMode, 
        sizeof(pSource->operatingMode)))
    {
        ERRORPRINT("Failed to set RedZoneOperatingModeProp to %d\n",
            pSource->operatingMode);
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZoneFrequencyKHzProp, &pSource->frequencyKHz, 
        sizeof(pSource->frequencyKHz)))
    {
        ERRORPRINT("Failed to set frequency %d KHz\n", pSource->frequencyKHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    RedZonePLPSet val;

    val.plp0_id = 0;
    val.plp1_id = 255;
    val.plp2_id = 255;
    val.plp3_id = 255;

    if (RedZoneCaptureSetProp(pSource->hRedZoneCapture, 
        RedZonePLPSelectionProp, &val, sizeof(val)))
    {
        ERRORPRINT("Failed to set PLP selection\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (RedZoneCaptureRegisterCallbacks(pSource->hRedZoneCapture,
        ATSC3BasebandRZRSDK2SourceBasebandPacketCallback, pSource))
    {
        ERRORPRINT("Failed to register baseband packet callback\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->pBasebandPayload = (uint8_t *)malloc(MAX_BASEBAND_PAYLOAD_SIZE);
    if (!pSource->pBasebandPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->fOpened = 1;
    pSource->basebandPacketNumber = 0;
    pSource->plpIdMask = 0;
    pSource->fPLPIdMaskChanged = 0;
    pSource->PLPsCallback = NULL;
    pSource->LLSPLPsCallback = NULL;
    pSource->BasebandPacketCallback = NULL;
    pSource->pCallbackUserData = NULL;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (pthread_mutex_unlock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;

errout:
    if (pSource->pBasebandPayload)
    {
        free(pSource->pBasebandPayload);
        pSource->pBasebandPayload = NULL;
    }

    goto out;
}

ENStatus
ATSC3BasebandRZRSDK2SourceClose(ENATSC3Source hSource)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    free(pSource->pBasebandPayload);
    pSource->pBasebandPayload = NULL;

    pSource->fOpened = 0;

    if (pthread_mutex_unlock(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandRZRSDK2SourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3BasebandRZRSDK2Source *pSource =
        (ENATSC3BasebandRZRSDK2Source *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->plpIdMask = plpIdMask;
    pSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandRZRSDK2SourceRegisterCallbacks(ENATSC3Source pSource,
    ENATSC3SourcePLPsCallbackFunction PLPsCallback,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback,
    ENATSC3SourceBasebandPacketCallbackFunction BasebandPacketCallback,
    void *pUserData)
{
    ENATSC3BasebandRZRSDK2Source *pBasebandRZRSDK2Source =
        (ENATSC3BasebandRZRSDK2Source *)pSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandRZRSDK2Source->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandRZRSDK2Source->PLPsCallback = PLPsCallback;
    pBasebandRZRSDK2Source->LLSPLPsCallback = LLSPLPsCallback;
    pBasebandRZRSDK2Source->BasebandPacketCallback = BasebandPacketCallback;
    pBasebandRZRSDK2Source->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pBasebandRZRSDK2Source->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceInitialize(int32_t *pNDevices)
{
    int32_t fHaveLock = 0;
    RedZoneCaptureHandle_t hRedZoneCapture = NULL;
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;

    if (pthread_mutex_init(&atsc3BasebandRZRSDK2SourceGlobalsG.lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
    atsc3BasebandRZRSDK2SourceGlobalsG.nSources = 0;

    while (1)
    {
        int32_t rzrret;

        if (atsc3BasebandRZRSDK2SourceGlobalsG.nSources ==
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources))
        {
            ERRORPRINT("Too many sources\n");
            break;
        }

        rzrret = RedZoneCaptureOpen(&hRedZoneCapture);
        if (rzrret)
        {
            ERRORPRINT("RedZoneCaptureOpen failed %d\n", rzrret);
            rval = ENStatus_Failure;
            goto errout;
        }

        rzrret = RedZoneCaptureInit(hRedZoneCapture);
        if (rzrret)
        {
            ERRORPRINT("Failed to init RedZoneCapture %d\n", rzrret);
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                ERRORPRINT("Failed to close RedZoneCapture\n");
            }
            rval = ENStatus_Failure;
            goto errout;
        }

        pSource->base.Start = ATSC3BasebandRZRSDK2SourceStart;
        pSource->base.Close = ATSC3BasebandRZRSDK2SourceClose;
        pSource->base.SetPLPIds = ATSC3BasebandRZRSDK2SourceSetPLPIds;
        pSource->base.RegisterCallbacks = 
            ATSC3BasebandRZRSDK2SourceRegisterCallbacks;
        pSource->hRedZoneCapture = hRedZoneCapture;
#if 0
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneModelNumberStringProp, pSource->base.strModel,
            sizeof(pSource->base.strModel) != RZR_SUCCESS))
        {
            ERRORPRINT("Failed to get model\n");
            pSource->base.strModel[0] = 0;
        }
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneSerialNumberStringProp, pSource->base.strSerial,
            sizeof(pSource->base.strSerial) != RZR_SUCCESS))
        {
            ERRORPRINT("Failed to get serial\n");
            pSource->base.strSerial[0] = 0;
        }
#else
        uint32_t model;
        if (RedZoneCaptureGetProp(hRedZoneCapture, 
            RedZoneModelNumberProp, &model,
            sizeof(model) != RZR_SUCCESS))
        {
            ERRORPRINT("Failed to get model\n");
            pSource->base.strModel[0] = 0;
        }
        else
        {
            snprintf(pSource->base.strModel, sizeof(pSource->base.strModel),
                "RZR-%d", model);
            pSource->base.strModel[sizeof(pSource->base.strModel) - 1] = 0;

        }
        pSource->base.strSerial[0] = 0;
#endif

        pSource->fOpened = 0;
        if (pthread_mutex_init(&pSource->lock, NULL))
        {
            ERRORPRINT("Failed to init lock\n");
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                ERRORPRINT("Failed to close RedZoneCapture\n");
            }
            rval = ENStatus_Failure;
            goto errout;
        }

        if (pthread_mutex_init(&pSource->statsLock, NULL))
        {
            ERRORPRINT("Failed to init lock\n");
            if (pthread_mutex_destroy(&pSource->lock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
            if (RedZoneCaptureClose(hRedZoneCapture))
            {
                ERRORPRINT("Failed to close RedZoneCapture\n");
            }
            rval = ENStatus_Failure;
            goto errout;
        }

        atsc3BasebandRZRSDK2SourceGlobalsG.nSources++;
        pSource++;

        break;
    }

    *pNDevices = atsc3BasebandRZRSDK2SourceGlobalsG.nSources;

out:
    return rval;

errout:
    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources;
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            ERRORPRINT("Failed to close RedZoneCapture\n");
            pSource->hRedZoneCapture = NULL;
        }
    }

    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    goto out;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceShutdown(void)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources; 
         i < atsc3BasebandRZRSDK2SourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (RedZoneCaptureClose(pSource->hRedZoneCapture))
        {
            ERRORPRINT("Failed to close RedZoneCapture\n");
        }
        pSource->hRedZoneCapture = NULL;

        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (pthread_mutex_destroy(&atsc3BasebandRZRSDK2SourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandRZRSDK2SourceGetDeviceInfo(int32_t deviceNumber,
    char *pModel, int32_t nModelBytes, char *pSerial, int32_t nSerialBytes)
{
    ENATSC3BasebandRZRSDK2Source *pSource;
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= atsc3BasebandRZRSDK2SourceGlobalsG.nSources)
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            numof(atsc3BasebandRZRSDK2SourceGlobalsG.vSources));
        rval = ENStatus_Failure;
        goto out;
    }

    pSource = atsc3BasebandRZRSDK2SourceGlobalsG.vSources + deviceNumber;
    strncpy(pModel, pSource->base.strModel, nModelBytes - 1);
    pModel[nModelBytes - 1] = 0;
    strncpy(pSerial, pSource->base.strSerial, nSerialBytes - 1);
    pSerial[nSerialBytes - 1] = 0;

out:
    return rval;
}
