/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>
#include <libenatsc3utils/enipfilterset.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>

#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>

#include <libenatsc3route/atsc3routesls_mime.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>

#include <libenatsc3dashfeeder/libenatsc3dashfeeder.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

struct UserData
{
    ENVerbosity verbosity;
    int32_t fIgnoreChecksumErrors;
    int ifd;

    EnIPFilterSet *pIPFilterSet;

    char *pOutPath;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;
    ENATSC3IPV4Handle_t hATSC3IPV4;
    ENATSC3ROUTESessionHandle_t hROUTESession;
    ENATSC3DASHFeederHandle_t hDASHFeeder;
    uint32_t destinationAddress;
    uint16_t destinationPort;

    uint64_t plpIds;
};

int32_t fRunningG = 1;

ENStatus MediaFileBeginCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, uint8_t codepoint, 
    const char *pId, ENATSC3ROUTESessionIdType idType, const char *pURL, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t fError,
    void *pUserData)
{
    ENStatus rval = ENStatus_Success;

    return rval;
}

void 
MediaFileEndCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t tsi, uint32_t toi, const char *pURL, int64_t objectPosition, 
    int64_t objectSize, void *pUserData)
{
    return;
}

void 
MediaFileDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t tsi, uint32_t toi, const char *pURL, const uint8_t *pPayload, 
    int32_t sPayload, int64_t objectPosition, int64_t objectSize, int32_t fError, 
    void *pUserData)
{
    return;
}

void 
MediaFileAbortCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    uint32_t tsi, uint32_t toi, const char *pURL, int64_t objectPosition, 
    int64_t objectSize, void *pUserData)
{
    return;
}

void 
LCTChannelRemovedCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, void *pUserData)
{
    return;
}

int32_t
HandleUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t plpId, uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, struct UserData *pUserData)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        fprintf(stderr, "UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength < length)
    {
        fprintf(stderr, "Not enough data for UDP packet: %d < %d\n",
            packetLength, length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    if (EnIPFilterSetFilter(destinationAddress, destPort,
        pUserData->pIPFilterSet))
    {
        goto out;
    }

    if (checksum)
    {
       struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader =
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        if (pUserData->verbosity > ENVerbosity_Pedantic)
        {
            printf("UDP\n");
            printf("  Source port: %d\n", sourcePort);
            printf("  Dest port: %d\n", destPort);
            printf("  Length: %d\n", length);
            printf("  Checksum: %x\n", checksum);
        }

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);
    }
    else
    {
        sum = 0;
    }

    if (sum == 0)
    {
        if (pUserData->verbosity > ENVerbosity_Pedantic)
        {
            printf("  Checksum verified\n");
        }
    }
    else
    {
        static int32_t fChecksumErrorLogged = 0;

        if (pUserData->verbosity >= ENVerbosity_Pedantic || 
            !fChecksumErrorLogged)
        {
            printf("  UDP CHECKSUM ERROR: %x%s\n", sum, 
                pUserData->fIgnoreChecksumErrors ? " (ignored)" : "");
            if (pUserData->verbosity < ENVerbosity_Pedantic)
            {
                printf("Further checksum errors suppressed; enable pedantic verbosity to view\n");
            }
            fChecksumErrorLogged = 1;
        }
    }

    if (sum == 0 || pUserData->fIgnoreChecksumErrors)
    {
        if (ENATSC3ROUTESessionProcessUDPPacket(pUserData->hROUTESession, 
            pPacketTimeUTC, packetNumber, plpId, pPacket + 8, packetLength - 8))
        {
            fprintf(stderr, "Failed to process UDP packet\n");
        }
    }

out:
    return rval;
}

void
TransportPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    switch (protocol)
    {
        case 17:
            HandleUDPPacket(pPacketTimeUTC, packetNumber, plpId, sourceAddress, 
                destinationAddress, pPacket, sPacket, pUserData);
            break;
        default:
            fprintf(stderr, "Unhandled protocol %d\n", protocol);
            break;
    }
}

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, pPacketTimeUTC, 
        packetNumber, pPayload, payloadLength);

    return;
}

int32_t
HandleLinkMappingTable(
    ENATSC3ALPLinkLayerSignalingHeader *pLinkLayerSignalingHeader, 
    uint32_t plpId, struct timespec *pPacketTimeUTC, const uint8_t *pPayload, 
    int32_t payloadLength, struct UserData *pUserData)
{
    uint32_t numPLPs;
    int i;
    int j;

    // Skip over signalling header
    pPayload += 5;

    numPLPs = (pPayload[0] >> 2) + 1;
    pPayload++;

    if (pUserData->verbosity > ENVerbosity_Pedantic)
    {
        printf("num PLPs: %d\n", numPLPs);

        for (i = 0; i < numPLPs; i++)
        {
            uint32_t plpID;
            uint32_t numMulticasts;

            plpID = pPayload[0] >> 2;
            numMulticasts = pPayload[1];
            pPayload += 2;

            printf("  plp: %d\n", plpID);
            printf("    multicasts: %d\n", numMulticasts);

            for (j = 0; j < numMulticasts; j++)
            {
                uint32_t srcIPAddr;
                uint32_t dstIPAddr;
                uint32_t srcUDPPort;
                uint32_t dstUDPPort;
                uint32_t sidFlag;
                uint32_t compressedFlag;
                uint32_t SID;
                uint32_t contextId;

                srcIPAddr = (pPayload[0] << 24) | (pPayload[1] << 16) | 
                    (pPayload[2] << 8) | pPayload[3]; 
                dstIPAddr = (pPayload[4] << 24) | (pPayload[5] << 16) | 
                    (pPayload[6] << 8) | pPayload[7]; 
                srcUDPPort = (pPayload[8] << 8) | pPayload[9];
                dstUDPPort = (pPayload[10] << 8) | pPayload[11];
                sidFlag = pPayload[12] & 0x80;
                compressedFlag = pPayload[12] & 0x40;

                pPayload += 13;

                if (sidFlag)
                {
                    SID = pPayload[0];
                    pPayload++;
                }

                if (compressedFlag)
                {
                    contextId = pPayload[0];
                    pPayload++;
                }

                printf("      Multicast %d\n", j);
                printf("        src: %d.%d.%d.%d:%d\n", 
                    (srcIPAddr >> 24)  & 0xff, (srcIPAddr >> 16)  & 0xff,
                    (srcIPAddr >> 8)  & 0xff, srcIPAddr  & 0xff,
                    srcUDPPort);
                printf("        dst: %d.%d.%d.%d:%d\n", 
                    (dstIPAddr >> 24)  & 0xff, (dstIPAddr >> 16)  & 0xff,
                    (dstIPAddr >> 8)  & 0xff, dstIPAddr  & 0xff,
                    dstUDPPort);
                printf("        SID flag: %d\n", sidFlag);
                if (sidFlag)
                {
                    printf("        SID: %d\n", SID);
                }
                printf("        compressed flag: %d\n", compressedFlag);
                if (compressedFlag)
                {
                    printf("        context id: %d\n", contextId);
                }
            }
        }
    }

    return 0;
}

void
LinkLayerSignalingPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPayload, 
    int32_t payloadLength, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;
    ENATSC3ALPLinkLayerSignalingHeader llsheader;

    llsheader.signaling_type = pPayload[0];
    llsheader.signaling_type_extension = (pPayload[1] << 8) | pPayload[2];
    llsheader.signaling_version = pPayload[3];
    llsheader.signaling_format = (pPayload[4] >> 6) & 0x3;
    llsheader.signaling_encoding = (pPayload[4] >> 4) & 0x3;

    switch (llsheader.signaling_type)
    {
        case ENATSC3ALPSignalingType_LinkMappingTable:
            HandleLinkMappingTable(&llsheader, plpId, pPacketTimeUTC, 
                pPayload, payloadLength, pUserData);
            break;
        default:
            printf("Unhandled link layer signaling type %d\n", 
                llsheader.signaling_type);
            break;
    }

    return;
}

void
FileReaderIPV4PacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3IPV4ProcessIPV4Packet(pUserData->hATSC3IPV4, plpId, 
        pPacketTimeUTC, packetNumber, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process IPV4 packet\n");
    }
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void
AddIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    printf("Adding %x:%d\n", addr, port);

    if (EnIPFilterSetAddFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to add filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
RemoveIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (EnIPFilterSetRemoveFilter(addr, port, pUserData->pIPFilterSet))
    {
        fprintf(stderr, "Failed to remove filter for %.d.%.d.%.d.%.d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
CommitIPAddresses(void *pCbUserData)
{
    // Nothing to do
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }

}

void
FileReaderBasebandPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, 
    const ENATCS3L1Time *pL1Time, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband,
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pL1Time, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutPath, int32_t *pTOffsetSeconds, 
    uint64_t *pPLPIds, uint32_t *pDestinationAddr, uint16_t *pDestinationPort, 
    ENVerbosity *pVerbosity, int32_t *pFIgnoreChecksumErrors)
{
    int32_t err = 0;
    int32_t fHaveDestinationAddr = 0;
    int32_t fHaveDestinationPort = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;
    *pPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-pcapipv4"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_PCAPIPV4;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-port"))
        {
            if (i + 1 < argc)
            {
                *pDestinationPort = atol(argv[i + 1]);
                fHaveDestinationPort = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-addr"))
        {
            if (i + 1 < argc)
            {
                struct in_addr addr;
                if (inet_aton(argv[i + 1], &addr))
                {
                    *pDestinationAddr = htonl(addr.s_addr);
                    fHaveDestinationAddr = 1;
                }
                else err = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbosity = atol(argv[i + 1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i + 1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignorechecksum"))
        {
            *pFIgnoreChecksumErrors = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutPath == NULL || 
        !fHaveDestinationAddr || !fHaveDestinationPort)
    {
        err =1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b alpcapturefile | -bb basebandcapturefile -o path -plp plpid [-plp plpid ...] -addr multicast_address -port multicast_port [-offsetms offset] [-offset seconds] [-v] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    struct UserData userdata = 
    {
        .verbosity = ENVerbosity_Terse,
        .ifd = -1,
        .fIgnoreChecksumErrors = 0,
        .hATSC3IPV4 = NULL,
        .hROUTESession = NULL,
        .pOutPath = NULL,
        .pIPFilterSet = NULL,
    };
    struct sigaction act;
    int32_t tOffsetSeconds = 0;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &userdata.pOutPath, 
        &tOffsetSeconds, &userdata.plpIds,
        &userdata.destinationAddress, &userdata.destinationPort, 
        &userdata.verbosity, &userdata.fIgnoreChecksumErrors))
    {
        rval = 1;
        goto out;
    }

    if (EnIPFilterSetAllocate(64, &userdata.pIPFilterSet))
    {
        fprintf(stderr, "Failed to allocate IP filter set\n");
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (inFileType != ENATSC3FileType_PCAPIPV4)
    {
        if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
        {
            fprintf(stderr, "Failed to open ALP processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
            LinkLayerSignalingPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
        {
            fprintf(stderr, "Failed to set ALP PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3IPV4Open(&userdata.hATSC3IPV4))
    {
        fprintf(stderr, "Failed to open ATSC3IPV4 session\n");
        rval = 1;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(userdata.hATSC3IPV4, userdata.verbosity);

    if (ENATSC3IPV4RegisterCallbacks(userdata.hATSC3IPV4,
        TransportPacketCallback, &userdata))
    {
        fprintf(stderr, "Failed to register ATSC3IPV4 callback\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType_ATSC3_DASH, 
        userdata.destinationAddress, userdata.destinationPort, 
        AddIPAddress, RemoveIPAddress, CommitIPAddresses, 
        &userdata, "test", &userdata.hROUTESession))
    {
        fprintf(stderr, "Failed to allocate ROUTE session\n");
        rval = 1;
        goto out;
    }

    ENATSC3ROUTESessionSetVerbosity(userdata.hROUTESession, userdata.verbosity);

    if (ENATSC3DASHFeederAllocate(userdata.hROUTESession, "test", 
        &userdata.hDASHFeeder))
    {
        fprintf(stderr, "Failed to allocate DASH feeder\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3DASHFeederRegisterCallbacks(
        userdata.hDASHFeeder,
        MediaFileBeginCallback,
        MediaFileEndCallback,
        MediaFileAbortCallback,
        MediaFileDataCallback,
        LCTChannelRemovedCallback,
        &userdata))
    {
        fprintf(stderr, "Failed to register DASH feeder callbacks\n");
        rval = 1;
        goto out;
    }

    ENATSC3DASHFeederSetVerbosity(userdata.hDASHFeeder, userdata.verbosity);

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        case ENATSC3FileType_PCAPIPV4:
            rval = ENATSC3FileReaderReadPCAPIPV4File(userdata.ifd, 
                tOffsetSeconds, FileReaderIPV4PacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (userdata.hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
        {
            fprintf(stderr, "Failed to clear ALP PLP mask\n");
            rval = 1;
            goto out;
        }
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0)) 
        {
            fprintf(stderr, "Failed to clear baseband PLP mask\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (userdata.hDASHFeeder)
    {
        if (ENATSC3DASHFeederFree(userdata.hDASHFeeder))
        {
            fprintf(stderr, "Failed to free DASH feeder\n");
        }
    }

    if (userdata.hROUTESession)
    {
        if (ENATSC3ROUTESessionFree(userdata.hROUTESession))
        {
            fprintf(stderr, "Failed to free ROUTE session\n");
        }
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (userdata.hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(userdata.hATSC3IPV4))
        {
            fprintf(stderr, "Failed to close ATSC3IPV4 session\n");
        }
    }

    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.pIPFilterSet)
    {
        if (EnIPFilterSetFree(userdata.pIPFilterSet))
        {
            fprintf(stderr, "Failed to free IP filter set\n");
        }
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
