/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <libenatsc3filereader/libenatsc3filereader.h>

#include <enatsc3cap/enudpcap.h>

#define MICROSECONDS_PER_SECOND 1000000

struct UserData
{
    int ifd;
    int ofd;
    int32_t verbose;

    ENATSC3BasebandHandle_t hATSC3Baseband;
    ENATSC3ALPHandle_t hATSC3ALP;

    uint64_t plpIds;

    // The timestamp of the last pcap record written
    uint32_t ts_sec;
    uint32_t ts_usec;
};

int32_t fRunningG = 1;

void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    struct UserData *pUserData = (struct UserData *)pCbData;

    struct 
    {
        uint32_t ts_sec;         /* timestamp seconds */
        uint32_t ts_usec;        /* timestamp microseconds */
        uint32_t incl_len;       /* number of octets of packet saved in file */
        uint32_t orig_len;       /* actual length of packet */
    } __attribute__((packed)) pcaprec_hdr;
    struct
    {
        uint8_t vDestinationMAC[6];
        uint8_t vSourceMAC[6];
        uint16_t typelen;
    } __attribute__((packed)) ethernetII_frame_hdr;
    uint8_t vSourceMAC[6] = {0x00, 0x0c, 0x29, 0x8a, 0xc9, 0x4f};
    uint8_t vDestinationMAC[6] = {0x00, 0x0c, 0x29, 0x8a, 0xc9, 0x50};

    if ((pPayload[0] & 0xf0) != 0x40)
    {
        fprintf(stderr, "Got unexpected ip version %d, skipping\n",
            pPayload[0] >> 4);
        goto out;
    }

    pcaprec_hdr.ts_sec = pPacketTimeUTC->tv_sec;
    pcaprec_hdr.ts_usec = pPacketTimeUTC->tv_nsec / 1000;

    if (pcaprec_hdr.ts_sec < pUserData->ts_sec ||
        (pcaprec_hdr.ts_sec == pUserData->ts_sec &&
         pcaprec_hdr.ts_usec <= pUserData->ts_usec))
    {
        pcaprec_hdr.ts_sec = pUserData->ts_sec;
        pcaprec_hdr.ts_usec = pUserData->ts_usec;
        pcaprec_hdr.ts_usec += 10;
        if (pcaprec_hdr.ts_usec >= MICROSECONDS_PER_SECOND)
        {
            pcaprec_hdr.ts_usec -= MICROSECONDS_PER_SECOND;
            pcaprec_hdr.ts_sec++;
        }
    }

    pUserData->ts_sec = pcaprec_hdr.ts_sec;
    pUserData->ts_usec = pcaprec_hdr.ts_usec;

    pcaprec_hdr.incl_len = payloadLength + sizeof(ethernetII_frame_hdr);
    pcaprec_hdr.orig_len = payloadLength + sizeof(ethernetII_frame_hdr);

    memcpy(ethernetII_frame_hdr.vDestinationMAC, vDestinationMAC, 6);
    memcpy(ethernetII_frame_hdr.vSourceMAC, vSourceMAC, 6);
    ethernetII_frame_hdr.typelen = 8;

    if (pUserData->verbose)
    {
        printf("%"PRId64".%09ld %d bytes\n", pPacketTimeUTC->tv_sec, 
            pPacketTimeUTC->tv_nsec, payloadLength);
    }

    if (write(pUserData->ofd, &pcaprec_hdr, sizeof(pcaprec_hdr)) != 
        sizeof(pcaprec_hdr))
    {
        fprintf(stderr, "Error writing pcap record header: %s\n", 
            strerror(errno));
        goto out;
    }

    if (write(pUserData->ofd, &ethernetII_frame_hdr, 
        sizeof(ethernetII_frame_hdr)) != sizeof(ethernetII_frame_hdr))
    {
        fprintf(stderr, "Error writing ethernet frame header: %s\n", 
            strerror(errno));
        goto out;
    }

    if (write(pUserData->ofd, pPayload, payloadLength) != payloadLength) 
    {
        fprintf(stderr, "Error writing ipv4 file: %s\n", 
            strerror(errno));
        goto out;
    }

out:
    return;
}

void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3FileType *pInFileType, char **ppOutFilename, 
    int32_t *pTOffsetSeconds, uint64_t *pPLPIds, int32_t *pVerbose)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutFilename = NULL;
    *pPLPIds = 0;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_Baseband;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-j288alp"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                *pInFileType = ENATSC3FileType_J288ALP;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-plp"))
        {
            if (i + 1 < argc)
            {
                *pPLPIds |= 1ULL << atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbose = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-offset"))
        {
            if (i + 1 < argc)
            {
                *pTOffsetSeconds = atol(argv[i+1]);
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (*ppInFilename == NULL || *ppOutFilename == NULL)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b alpcapturefile | -bb basebandcapturefile -o pcapfile -plp plpid [-plp plpid ...] [-offset seconds] [-v level] [-ignorechecksum]\n", argv[0]);
        err = 1;
    }

    return err;
}

int32_t
WritePCAPGlobalHeader(int ofd)
{
    int32_t rval = 0;

    struct 
    {
        uint32_t magic_number;  /* magic number */
        uint16_t version_major; /* major version number */
        uint16_t version_minor; /* minor version number */
        int32_t  thiszone;      /* GMT to local correction */
        uint32_t sigfigs;       /* accuracy of timestamps */
        uint32_t snaplen;       /* max length of captured packets, in octets */
        uint32_t network;       /* data link type */
    } __attribute__((packed)) globalHeader = 
    {
        .magic_number = 0xa1b2c3d4,
        .version_major = 2,
        .version_minor = 4,
        .thiszone = 0,
        .sigfigs = 0,
        .snaplen = 0x00040000,
        .network = 1
    };

    if (write(ofd, &globalHeader, sizeof(globalHeader)) != sizeof(globalHeader))
    {
        fprintf(stderr, "Failed to write PCAP global header\n");
        rval = 1;
        goto out;
    }

out:
    return rval;
}

void
FileReaderALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3ALPProcessALPPacket(pUserData->hATSC3ALP, plpId, pPacketTimeUTC,
        packetNumber, fError, fDiscontinuity, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process ALP packet\n");
    }
}

void 
FileReaderBasebandPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, 
    const ENATCS3L1Time *pL1Time, const uint8_t *pPacket, int32_t 
    sPacket, void *pCbUserData)
{
    struct UserData *pUserData = (struct UserData *)pCbUserData;

    if (ENATSC3BasebandProcessBasebandPacket(pUserData->hATSC3Baseband, 
        plpId, pPacketTimeUTC, packetNumber, fError, fDiscontinuity, 
        pL1Time, pPacket, sPacket))
    {
        fprintf(stderr, "Failed to process baseband packet\n");
    }
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        fRunningG = 0;
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    char *pInFilename;
    ENATSC3FileType inFileType;
    char *pOutFilename;
    int32_t tOffsetSeconds = 0;
    struct UserData userdata = 
    {
        .ifd = -1,
        .ofd = -1,
        .verbose = 0,
        .hATSC3Baseband = NULL,
        .hATSC3ALP = NULL,
    };
    struct sigaction act;
    int rval = 0;

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &inFileType, &pOutFilename, 
        &tOffsetSeconds, &userdata.plpIds, &userdata.verbose))
    {
        rval = 1;
        goto out;
    }

    if (inFileType == ENATSC3FileType_Baseband)
    {
        if (ENATSC3BasebandOpen(&userdata.hATSC3Baseband))
        {
            fprintf(stderr, "Failed to open baseband processor\n");
            rval = 1;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(userdata.hATSC3Baseband,
            ALPPacketCallback, &userdata))
        {
            fprintf(stderr, "Failed to register baseband callbacks\n");
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPOpen(&userdata.hATSC3ALP))
    {
        fprintf(stderr, "Failed to open ALP processor\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3ALPRegisterCallbacks(userdata.hATSC3ALP, IPV4PacketCallback, 
        NULL, &userdata))
    {
        fprintf(stderr, "Failed to register callbacks\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband,
            userdata.plpIds))
        {
            fprintf(stderr, "Failed to set baseband PLP mask %"PRIx64"\n",
                userdata.plpIds);
            rval = 1;
            goto out;
        }
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, userdata.plpIds))
    {
        fprintf(stderr, "Failed to set ALP PLP mask %"PRIx64"\n", 
            userdata.plpIds);
        rval = 1;
        goto out;
    }

    userdata.ifd = open(pInFilename, O_RDONLY);
    if (userdata.ifd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", 
            pInFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    userdata.ofd = open(pOutFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (userdata.ofd < 0)
    {
        fprintf(stderr, "Failed to open %s for writing: %s\n",
            pOutFilename, strerror(errno));
        rval = 1;
        goto out;
    }

    if (WritePCAPGlobalHeader(userdata.ofd))
    {
        rval = 1;
        goto out;
    }

    switch (inFileType)
    {
        case ENATSC3FileType_ALP:
            rval = ENATSC3FileReaderReadALPFile(userdata.ifd, tOffsetSeconds, 
                FileReaderALPPacketCallback, NULL, &fRunningG, &userdata);
            break;
        case ENATSC3FileType_Baseband:
            rval = ENATSC3FileReaderReadBasebandFile(userdata.ifd, 
                tOffsetSeconds, FileReaderBasebandPacketCallback, NULL, 
                &fRunningG, &userdata);
            break;
        case ENATSC3FileType_J288ALP:
            rval = ENATSC3FileReaderReadJ288ALPFile(userdata.ifd, 
                tOffsetSeconds, FileReaderALPPacketCallback, NULL, &fRunningG, 
                &userdata);
            break;
        default:
            fprintf(stderr, "Unhandled input file type %d\n", inFileType);
            rval = 1;
            goto out;
            break;
    }

    if (ENATSC3ALPSetPLPs(userdata.hATSC3ALP, 0))
    {
        fprintf(stderr, "Failed to clear ALP PLP mask\n");
        rval = 1;
        goto out;
    }

    if (userdata.hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(userdata.hATSC3Baseband, 0))
        {
            fprintf(stderr, "Failed to clear baseband PLP mask\n");
            rval = 1;
            goto out;
        }
    }
    
out:
    if (userdata.hATSC3ALP)
    {
        ENATSC3ALPClose(userdata.hATSC3ALP);
    }

    if (userdata.hATSC3Baseband)
    {
        ENATSC3BasebandClose(userdata.hATSC3Baseband);
    }

    if (userdata.ofd >= 0)
    {
        close(userdata.ofd);
    }

    if (userdata.ifd >= 0)
    {
        close(userdata.ifd);
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
