/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 * 
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>

#include "libenatsc3dashfeeder.h"

static const char *ensys = "dashfdr";
static const char *ensubsys = "main";

#define LCT_MAX_OBJECT_SIZE (16 * 1024 * 1024)

#define REP_ID_MAX_LENGTH 256

#define DASHCONSUMER_URL_MAX_LEN 256

typedef struct _ENATSC3DASHFeeder
{
    ENVerbosity verbosity;

    // The ROUTE session this dashconsumer received data from.
    ENATSC3ROUTESessionHandle_t hROUTESession;

    char strDebugName[16];

    struct _LCTChannelInfo
    {
        int32_t fInUse;

        uint32_t tsi;
        char strId[REP_ID_MAX_LENGTH];
        ENATSC3ROUTESessionIdType idType;

        // An LCT channel can simultaneously deliver multiple objects. We
        // support receiving numof(vLCTChannelObjects) simutleously. Since
        // incomplete objects may be received, an object will be aborted if
        // we run out of vLCTChanneObject slots.
        struct _LCTChannelObject
        {
            int32_t fInUse;
            struct timespec tAllocated;

            uint64_t toi;
            uint8_t codepoint;
            char strURL[DASHCONSUMER_URL_MAX_LEN];
            int64_t objectSize;
            int64_t pos;
            int32_t payloadOffset;

            // fRejected is set if the application refuses the object in the
            // Accept callback.
            int64_t fRejected;

            // Applications can set a pointer in the WillAccept callback that 
            // will then be passed to the Begin, End, Abort, and Data 
            // callbacks.
            void *pObjectUserData;
        } vLCTChannelObjects[1];
    } vLCTChannelInfos[32];

    ENATSC3DASHFeederMediaFileBeginCallback MediaFileBeginCallback;
    ENATSC3DASHFeederMediaFileEndCallback MediaFileEndCallback;
    ENATSC3DASHFeederMediaFileAbortCallback MediaFileAbortCallback;
    ENATSC3DASHFeederMediaFileDataCallback MediaFileDataCallback;
    ENATSC3DASHFeederLCTChannelRemovedCallback LCTChannelRemovedCallback;
    void *pCallbackUserData;
} ENATSC3DASHFeeder;

static ENStatus
FindLCTChannelForTSI(ENATSC3DASHFeeder *pENATSC3DASHFeeder, uint32_t tsi,
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pENATSC3DASHFeeder->vLCTChannelInfos); i++)
    {
        if (pENATSC3DASHFeeder->vLCTChannelInfos[i].fInUse &&
            pENATSC3DASHFeeder->vLCTChannelInfos[i].tsi == tsi)
        {
            *ppLCTChannelInfo = pENATSC3DASHFeeder->vLCTChannelInfos + i;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

static ENStatus
AddLCTChannelForTSI(ENATSC3DASHFeeder *pENATSC3DASHFeeder, uint32_t tsi,
    const char *pId, ENATSC3ROUTESessionIdType idType, 
    struct _LCTChannelInfo **ppLCTChannelInfo)
{
    struct _LCTChannelInfo *pLCTChannelInfo;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
    {
        LOGPRINT_NOFILE("%s: Add LCT channel for TSI %u (%s)\n", 
            pENATSC3DASHFeeder->strDebugName, tsi, pId);
    }

    if (strlen(pId) >= sizeof(pENATSC3DASHFeeder->vLCTChannelInfos[0].strId))
    {
        ERRORPRINT("%s:%u id %s too long\n", pENATSC3DASHFeeder->strDebugName, 
            tsi, pId);
        goto out;
    }

    for (i = 0, pLCTChannelInfo = pENATSC3DASHFeeder->vLCTChannelInfos; 
         i < numof(pENATSC3DASHFeeder->vLCTChannelInfos); 
         i++, pLCTChannelInfo++)
    {
        if (!pENATSC3DASHFeeder->vLCTChannelInfos[i].fInUse)
        {
            struct _LCTChannelObject *pLCTChannelObject;
            int32_t j;

            pLCTChannelInfo->fInUse = 1;
            pLCTChannelInfo->tsi = tsi;
            strcpy(pLCTChannelInfo->strId, pId);
            pLCTChannelInfo->idType = idType;

            for (j = 0, pLCTChannelObject = pLCTChannelInfo->vLCTChannelObjects;
                 j < numof(pLCTChannelInfo->vLCTChannelObjects);
                 j++)
            {
                pLCTChannelObject->fInUse = 0;
            }
            *ppLCTChannelInfo = pLCTChannelInfo;
            rval = ENStatus_Success;
            break;
        }
    }

    if (i == numof(pENATSC3DASHFeeder->vLCTChannelInfos))
    {
        ERRORPRINT("%s: All LCT channels in use:\n", 
            pENATSC3DASHFeeder->strDebugName);
        for (i = 0; i < numof(pENATSC3DASHFeeder->vLCTChannelInfos); i++)
        {
            ERRORPRINT("  %d: fInUse %d tsi %u id %s type %d\n",
                i, pENATSC3DASHFeeder->vLCTChannelInfos[i].fInUse,
                pENATSC3DASHFeeder->vLCTChannelInfos[i].tsi,
                pENATSC3DASHFeeder->vLCTChannelInfos[i].strId);
        }
    }

out:
    return rval;
}

static ENStatus
RemoveLCTChannelForTSI(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    ENATSC3DASHFeeder *pENATSC3DASHFeeder, uint32_t tsi)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
    {
        LOGPRINT_NOFILE("%s: REMOVING LCT CHANNEL FOR TSI %u\n", 
            pENATSC3DASHFeeder->strDebugName, tsi);
    }

    pENATSC3DASHFeeder->LCTChannelRemovedCallback(pPacketTimeUTC, packetNumber, 
        tsi, pENATSC3DASHFeeder->pCallbackUserData);

    for (i = 0; i < numof(pENATSC3DASHFeeder->vLCTChannelInfos); i++)
    {
        if (pENATSC3DASHFeeder->vLCTChannelInfos[i].fInUse &&
            pENATSC3DASHFeeder->vLCTChannelInfos[i].tsi == tsi)
        {
            pENATSC3DASHFeeder->vLCTChannelInfos[i].fInUse = 0;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

void
ENATSC3ROUTEROUTELCTChannelAdded(struct timespec *pPacketTime, 
    uint64_t packetNumber, uint32_t addr, uint16_t port, uint32_t tsi, 
    const char *pId, ENATSC3ROUTESessionIdType idType, void *pCbUserData)
{
    struct _LCTChannelInfo *pLCTChannelInfo;
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = (ENATSC3DASHFeeder *)pCbUserData;

    if (AddLCTChannelForTSI(pENATSC3DASHFeeder, tsi, pId, idType, &pLCTChannelInfo))
    {
        ERRORPRINT("%s:%u Failed to allocate LCT Channel info\n", 
            pENATSC3DASHFeeder->strDebugName, tsi);
    }
}

void
ENATSC3DASHFeederROUTELCTChannelRemoved(struct timespec *pPacketTime, 
    uint64_t packetNumber, uint32_t addr, uint16_t port, uint32_t tsi, 
    void *pCbUserData)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = (ENATSC3DASHFeeder *)pCbUserData;

    if (RemoveLCTChannelForTSI(pPacketTime, packetNumber, pENATSC3DASHFeeder, 
        tsi))
    {
        ERRORPRINT("%s:%u Failed to free LCT Channel info\n", 
            pENATSC3DASHFeeder->strDebugName, tsi);
    }
}

static ENStatus
LCTChannelObjectAllocate(ENATSC3DASHFeeder *pENATSC3DASHFeeder, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    struct _LCTChannelInfo *pLCTChannelInfo, uint32_t toi, uint8_t codepoint, 
    struct _LCTChannelObject **ppLCTChannelObject)
{
    struct timespec tAllocated = {.tv_sec = 0x7fffffffffffffff, .tv_sec = 0};
    struct _LCTChannelObject *pLCTChannelObject;
    struct _LCTChannelObject *pOldestLCTChannelObject = NULL;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    *ppLCTChannelObject = NULL;

    for (i = 0, pLCTChannelObject = pLCTChannelInfo->vLCTChannelObjects; 
         i < numof(pLCTChannelInfo->vLCTChannelObjects); 
         i++, pLCTChannelObject++)
    {
        if (!pLCTChannelObject->fInUse)
        {
            break;
        }

        if ((tAllocated.tv_sec < 
             pLCTChannelObject->tAllocated.tv_sec) ||
            (tAllocated.tv_sec == pLCTChannelObject->tAllocated.tv_sec &&
             tAllocated.tv_nsec < pLCTChannelObject->tAllocated.tv_nsec))
        {
            tAllocated = pLCTChannelObject->tAllocated;
            pOldestLCTChannelObject = pLCTChannelObject;
        }
    }

    if (i == numof(pLCTChannelInfo->vLCTChannelObjects))
    {
        ENUtils_CrashAssert(pOldestLCTChannelObject);

        pLCTChannelObject = pOldestLCTChannelObject;
        if (!pLCTChannelObject->fRejected)
        {
            pENATSC3DASHFeeder->MediaFileAbortCallback(pPacketTimeUTC, 
                packetNumber, pLCTChannelInfo->tsi, toi, 
                pLCTChannelObject->strURL, pLCTChannelObject->pos,
                pLCTChannelObject->objectSize, 
               pENATSC3DASHFeeder->pCallbackUserData);
        }
    }

   memset(pLCTChannelObject, 0, sizeof(*pLCTChannelObject));
   pLCTChannelObject->fInUse = 1;
   clock_gettime(CLOCK_MONOTONIC, &pLCTChannelObject->tAllocated);
   pLCTChannelObject->toi = toi;
   pLCTChannelObject->codepoint = codepoint;
   pLCTChannelObject->objectSize = -1;

   *ppLCTChannelObject = pLCTChannelObject;

   return rval;
}

static void
LCTChannelObjectFree(struct _LCTChannelObject *pLCTChannelObject)
{
    pLCTChannelObject->fInUse = 0;
}

void
ENATSC3DASHFeederROUTEObjectDataCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, uint8_t codepoint, 
    int32_t formatId, int32_t frag, int32_t order, const char *pId, 
    ENATSC3ROUTESessionIdType idType, const uint8_t *pPayload, 
    int32_t payloadLength, int64_t objectPosition, int64_t tolTransferLength,
    int64_t ftiTransferLength, int32_t fError, void *pCbUserData)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = (ENATSC3DASHFeeder *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;
    struct _LCTChannelObject *pLCTChannelObject = NULL;
    int32_t i;
    int32_t fDrop = 0;
    int32_t payloadOffset = 0;

    if (FindLCTChannelForTSI(pENATSC3DASHFeeder, tsi, &pLCTChannelInfo))
    {
        ERRORPRINT("%s:%u Failed to find LCT channel info\n", 
            pENATSC3DASHFeeder->strDebugName, tsi);
        goto out;
    }

    for (i = 0; i < numof(pLCTChannelInfo->vLCTChannelObjects); i++)
    {
        if (pLCTChannelInfo->vLCTChannelObjects[i].fInUse && 
            pLCTChannelInfo->vLCTChannelObjects[i].toi == toi)
        {
            pLCTChannelObject = pLCTChannelInfo->vLCTChannelObjects + i;
            break;
        }
    }

    // See if this is a new object
    if (!pLCTChannelObject)
    {
        // We only start processing a new object at offset 0.
        if (objectPosition != 0)
        {
            if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT_NOFILE(
                    "%s:%u: WAITING FOR offset 0 (toi %u pos %"PRId64")\n", 
                    pENATSC3DASHFeeder->strDebugName, tsi, toi, objectPosition);
            }
            goto out;
        }
        else
        {
            if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT_NOFILE("%s:%u Starting toi %u\n", 
                    pENATSC3DASHFeeder->strDebugName, tsi, toi);
            }
        }
    }
    else if (pLCTChannelObject->fRejected)
    {
        if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT("%s:%u Rejected TOI %u\n", 
                pENATSC3DASHFeeder->strDebugName, tsi, toi);
        }
        goto out;
    }

    if (!pLCTChannelObject)
    {
        LCTChannelObjectAllocate(pENATSC3DASHFeeder, pPacketTimeUTC, 
            packetNumber, pLCTChannelInfo, toi, codepoint, &pLCTChannelObject);

        if (formatId == ENATSC3LCTChannelObjectFormatId_Entity)
        {
            if (ENATSC3ROUTESessionGetEntityModeObjectURLAndContentType(
                pENATSC3DASHFeeder->hROUTESession, tsi, pPayload, 
                payloadLength, &payloadOffset, pLCTChannelObject->strURL,
                sizeof(pLCTChannelObject->strURL), NULL, 0))
            {
                ERRORPRINT("Failed to get url for tsi %u toi %u\n",
                    tsi, toi);
                goto out;
            }

            pPayload += payloadOffset;
            payloadLength -= payloadOffset;
        }
        else if (formatId == ENATSC3LCTChannelObjectFormatId_File)
        {
            if (ENATSC3ROUTESessionGetObjectURLAndContentType(
                pENATSC3DASHFeeder->hROUTESession, tsi, toi, 
                pLCTChannelObject->strURL, sizeof(pLCTChannelObject->strURL), 
                NULL, 0))
            {
                ERRORPRINT("%s:%u Failed to get url for toi %u\n",
                    pENATSC3DASHFeeder->strDebugName, tsi, toi);
                goto out;
            }
        }

        if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT_NOFILE(
                "%s:%u %"PRId64".%06lu MediaFileBegin %s codepoint %u\n",
                pENATSC3DASHFeeder->strDebugName, tsi, pPacketTimeUTC->tv_sec, 
                pPacketTimeUTC->tv_nsec / 1000, pLCTChannelObject->strURL, 
                codepoint);
        }

        if (pENATSC3DASHFeeder->MediaFileBeginCallback(pPacketTimeUTC, 
            packetNumber, tsi, toi, codepoint, pId, idType, 
            pLCTChannelObject->strURL, pPayload, payloadLength, fError, 
            pENATSC3DASHFeeder->pCallbackUserData))
        {
            if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
            {
                LOGPRINT_NOFILE("%s:%u REJECTING TOI %d\n", 
                    pENATSC3DASHFeeder->strDebugName, tsi, toi);
            }
            pLCTChannelObject->fRejected = 1;
            LCTChannelObjectFree(pLCTChannelObject);
            goto out;
        }

        pLCTChannelObject->payloadOffset = payloadOffset;
    }

    if (pLCTChannelObject->objectSize == -1)
    {
        if (tolTransferLength != -1)
        {
            if (tolTransferLength > LCT_MAX_OBJECT_SIZE)
            {
                if (pENATSC3DASHFeeder->verbosity >= 
                    ENVerbosity_StreamEventTrace)
                {
                    LOGPRINT("%s:%u Rejecting LCT object with tolTransferLength %"PRId64" > %"PRId64"\n",
                        pENATSC3DASHFeeder->strDebugName, tsi, 
                        tolTransferLength, LCT_MAX_OBJECT_SIZE);
                }
                pLCTChannelObject->fRejected = 1;
            }
            else
            {
                pLCTChannelObject->objectSize = tolTransferLength;
            }
        }
        else if (ftiTransferLength != -1)
        {
            if (ftiTransferLength > LCT_MAX_OBJECT_SIZE)
            {
                if (pENATSC3DASHFeeder->verbosity >= 
                    ENVerbosity_StreamEventTrace)
                {
                    LOGPRINT("%s:%u Rejecting LCT object with ftiTransferLength %"PRId64" > %"PRId64"\n",
                        pENATSC3DASHFeeder->strDebugName, tsi, 
                        tolTransferLength, LCT_MAX_OBJECT_SIZE);
                }
                pLCTChannelObject->fRejected = 1;
            }
            else
            {
                pLCTChannelObject->objectSize = ftiTransferLength;
            }
        }
    }
    else
    {
        if (tolTransferLength != -1)
        {
            if (pLCTChannelObject->objectSize != tolTransferLength)
            {
                if (pENATSC3DASHFeeder->verbosity >= 
                    ENVerbosity_StreamEventTrace)
                {
                    LOGPRINT("%s:%u %"PRIu64" TOL object size changed %"PRId64" != %"PRId64"\n",
                        pENATSC3DASHFeeder->strDebugName, tsi, packetNumber, 
                        pLCTChannelObject->objectSize, tolTransferLength);
                }
                if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
                {
                    LOGPRINT_NOFILE("%s:%u MediaFileAbort %s\n",
                        pENATSC3DASHFeeder->strDebugName, tsi, 
                        pLCTChannelObject->strURL);
                }
                pENATSC3DASHFeeder->MediaFileAbortCallback(pPacketTimeUTC, 
                    packetNumber, tsi, toi, pLCTChannelObject->strURL,
                    pLCTChannelObject->pos, pLCTChannelObject->objectSize, 
                    pENATSC3DASHFeeder->pCallbackUserData);
                LCTChannelObjectFree(pLCTChannelObject);
                goto out;
            }
        }

        if (ftiTransferLength != -1)
        {
            if (pLCTChannelObject->objectSize != ftiTransferLength)
            {
                if (pENATSC3DASHFeeder->verbosity >= 
                    ENVerbosity_StreamEventTrace)
                {
                    LOGPRINT("%s:%u %"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n",
                        pENATSC3DASHFeeder->strDebugName, tsi, packetNumber, 
                        pLCTChannelObject->objectSize, ftiTransferLength);
                }
                if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
                {
                    LOGPRINT_NOFILE("%s:%u MediaFileAbort %s\n",
                        pENATSC3DASHFeeder->strDebugName, tsi,
                        pLCTChannelObject->strURL); 
                }
                LOGPRINT_NOFILE("ABORT\n");
                pENATSC3DASHFeeder->MediaFileAbortCallback(pPacketTimeUTC, 
                    packetNumber, tsi, toi, pLCTChannelObject->strURL, 
                    pLCTChannelObject->pos, pLCTChannelObject->objectSize, 
                    pENATSC3DASHFeeder->pCallbackUserData);
                LCTChannelObjectFree(pLCTChannelObject);
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the
    // SrcFlow's Payload element.
    if (pLCTChannelObject->pos != objectPosition)
    {
        if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT_NOCONSOLE("%s:%u %"PRIu64" %"PRIu64".%.06lu av offset error %"PRId64" != %"PRId64" (%"PRId64" missing) (%s size %"PRId64")\n",
                pENATSC3DASHFeeder->strDebugName, tsi, packetNumber,
                pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
                pLCTChannelObject->pos, objectPosition,
                objectPosition - pLCTChannelObject->pos,
                pLCTChannelObject->strURL, 
                pLCTChannelObject->objectSize);
        }

        if (pLCTChannelObject->pos < objectPosition)
        {
            // We can only pad the packet if we know the object size. We
            // cannot rely on fecPayloadId being correct since we have no
            // guarantee on the integrity of the packet data (UDP checksums
            // would provide this, but UDP checksums are not required and even
            // when they appear to be present in many cases the checksum is a
            // garbage value and cannot be relied on.)
            if (pLCTChannelObject->objectSize != -1)
            {
                if (objectPosition + payloadLength >
                    pLCTChannelObject->objectSize)
                {
                    if (pENATSC3DASHFeeder->verbosity >= 
                        ENVerbosity_ObjectTrace)
                    {
                        LOGPRINT_NOCONSOLE(
                            "%s:%u payload beyond segment end, dropping\n",
                            pENATSC3DASHFeeder->strDebugName, tsi);
                    }
                    fDrop = 1;
                }
                else
                {
                    if (pENATSC3DASHFeeder->verbosity >= 
                        ENVerbosity_ObjectTrace)
                    {
                        LOGPRINT_NOFILE("%s:%u PADDING MEDIA DATA (object position %"PRId64" lct pos %"PRId64" objectSize %"PRId64"\n",
                            pENATSC3DASHFeeder->strDebugName, tsi,
                            objectPosition, pLCTChannelObject->pos,
                            pLCTChannelObject->objectSize); 
                    }

                    // We should never be padding at the beginning of an 
                    // object, which is the only time we could have a 
                    // non-zero payloadOffset.
                    ENUtils_CrashAssert(payloadOffset == 0);
                    ENUtils_CrashAssert(pLCTChannelObject->pos >= 
                        pLCTChannelObject->payloadOffset);
                    ENUtils_CrashAssert(pLCTChannelObject->objectSize >= 
                        pLCTChannelObject->payloadOffset);

                    pENATSC3DASHFeeder->MediaFileDataCallback(pPacketTimeUTC, 
                        packetNumber, tsi, toi, pLCTChannelObject->strURL, 
                        NULL, objectPosition - pLCTChannelObject->pos, 
                        pLCTChannelObject->pos ? 
                            pLCTChannelObject->pos - 
                                pLCTChannelObject->payloadOffset :
                            0, 
                        pLCTChannelObject->objectSize == -1 ?
                            -1 :
                            pLCTChannelObject->objectSize - 
                                pLCTChannelObject->payloadOffset,
                        1, pENATSC3DASHFeeder->pCallbackUserData); 
                    pLCTChannelObject->pos +=
                        objectPosition - pLCTChannelObject->pos;
                }
            }
            else
            {
                ERRORPRINT("%s:%u Segment size not known - dropping packet\n",
                    pENATSC3DASHFeeder->strDebugName, tsi);
                fDrop = 1;
                goto out;
            }
        }
        else
        {
            ERRORPRINT("%s:%u Out of order packet, dropping\n",
                pENATSC3DASHFeeder->strDebugName, tsi);
            fDrop = 1;
        }
    }

    if (!fDrop)
    {
        ENUtils_CrashAssert(pLCTChannelObject->objectSize == -1 || 
            pLCTChannelObject->objectSize >= pLCTChannelObject->payloadOffset);
        pENATSC3DASHFeeder->MediaFileDataCallback(pPacketTimeUTC, packetNumber,
            tsi, toi, pLCTChannelObject->strURL, pPayload, payloadLength, 
            pLCTChannelObject->pos ? 
                pLCTChannelObject->pos - pLCTChannelObject->payloadOffset : 0, 
            pLCTChannelObject->objectSize == -1 ?  -1 : 
                pLCTChannelObject->objectSize - 
                pLCTChannelObject->payloadOffset, 0,
            pENATSC3DASHFeeder->pCallbackUserData);
        pLCTChannelObject->pos += payloadLength + payloadOffset;
    }

    // See if the object is now complete
    if (pLCTChannelObject->pos == pLCTChannelObject->objectSize)
    {
        if (pENATSC3DASHFeeder->verbosity >= ENVerbosity_ObjectTrace)
        {
            LOGPRINT_NOFILE("%s:%u MediaFileEnd %s\n", 
                pENATSC3DASHFeeder->strDebugName, tsi, 
                pLCTChannelObject->strURL);
        }
        pENATSC3DASHFeeder->MediaFileEndCallback(pPacketTimeUTC, packetNumber, 
            tsi, toi, pLCTChannelObject->strURL, pLCTChannelObject->objectSize, 
            pLCTChannelObject->pos, pENATSC3DASHFeeder->pCallbackUserData);
        LCTChannelObjectFree(pLCTChannelObject);
    }

out:
    return;
}

void
ENATSC3DASHFeederROUTESessionResetCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, void *pCbUserData)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = (ENATSC3DASHFeeder *)pCbUserData;

    LOGPRINT_NOCONSOLE("%s: ROUTE session reset\n", 
        pENATSC3DASHFeeder->strDebugName);
}

ENStatus
ENATSC3DASHFeederAllocate(ENATSC3ROUTESessionHandle_t hROUTESession, 
    const char *pDebugName, ENATSC3DASHFeederHandle_t *phENATSC3DASHFeeder)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = NULL;
    int32_t fRegisteredROUTELCTChannelCallbacks = 0;
    int32_t fROUTEObjectCallbacksRegistered = 0;
    ENStatus rval = ENStatus_Success;

    pENATSC3DASHFeeder = 
        (ENATSC3DASHFeeder *)calloc(sizeof(ENATSC3DASHFeeder), 1);
    if (!pENATSC3DASHFeeder)
    {
        ERRORPRINT("%s: Out of memory\n", pDebugName);
        rval = ENStatus_Failure;
        goto errout;
    }

    strncpy(pENATSC3DASHFeeder->strDebugName, pDebugName, 
        sizeof(pENATSC3DASHFeeder->strDebugName) - 1);
    pENATSC3DASHFeeder->strDebugName[sizeof(pENATSC3DASHFeeder->strDebugName) - 1] = 0;

    pENATSC3DASHFeeder->hROUTESession = hROUTESession;

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(hROUTESession, 
        ENATSC3ROUTEROUTELCTChannelAdded, 
        ENATSC3DASHFeederROUTELCTChannelRemoved, pENATSC3DASHFeeder))
    {
        ERRORPRINT("%s: Failed to register ROUTE LCT Channel callbacks\n",
            pENATSC3DASHFeeder->strDebugName);
        rval = ENStatus_Failure;
        goto errout;
    }
    fRegisteredROUTELCTChannelCallbacks = 1;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(hROUTESession,
        ENATSC3DASHFeederROUTEObjectDataCallback, 
        ENATSC3DASHFeederROUTESessionResetCallback, pENATSC3DASHFeeder))
    {
        ERRORPRINT("%s: Failed to register ROUTE object callbacks\n",
            pENATSC3DASHFeeder->strDebugName);
        rval = ENStatus_Failure;
        goto errout;
    }
    fROUTEObjectCallbacksRegistered = 1;

    *phENATSC3DASHFeeder = pENATSC3DASHFeeder;

out:
    return rval;

errout:
    if (fROUTEObjectCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterObjectCallbacks(hROUTESession,
            ENATSC3DASHFeederROUTEObjectDataCallback, 
            ENATSC3DASHFeederROUTESessionResetCallback, pENATSC3DASHFeeder))
        {
            ERRORPRINT(
                "%s: WARN: Failed to unregister ROUTE object callbacks\n",
                pENATSC3DASHFeeder->strDebugName);
        }
    }

    if (fRegisteredROUTELCTChannelCallbacks)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(hROUTESession, 
            ENATSC3ROUTEROUTELCTChannelAdded, ENATSC3DASHFeederROUTELCTChannelRemoved, 
            pENATSC3DASHFeeder))
        {
            ERRORPRINT(
                "%s: WARN: Failed to unregister ROUTE LCT Channel callbacks\n",
                pENATSC3DASHFeeder->strDebugName);
        }
    }

    if (pENATSC3DASHFeeder)
    {
        free(pENATSC3DASHFeeder);
    }

    goto out;
}

ENStatus
ENATSC3DASHFeederFree(ENATSC3DASHFeederHandle_t hENATSC3DASHFeeder)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = (ENATSC3DASHFeeder *)hENATSC3DASHFeeder;
    ENStatus rval = ENStatus_Success;

    if (ENATSC3ROUTESessionUnregisterObjectCallbacks(
        pENATSC3DASHFeeder->hROUTESession,
        ENATSC3DASHFeederROUTEObjectDataCallback, 
        ENATSC3DASHFeederROUTESessionResetCallback, pENATSC3DASHFeeder))
    {
        ERRORPRINT("%s: WARN: Failed to unregister ROUTE object callbacks\n",
            pENATSC3DASHFeeder->strDebugName);
    }

    if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
        pENATSC3DASHFeeder->hROUTESession, ENATSC3ROUTEROUTELCTChannelAdded, 
        ENATSC3DASHFeederROUTELCTChannelRemoved, pENATSC3DASHFeeder))
    {
        ERRORPRINT(
            "%s: WARN: Failed to unregister ROUTE LCT Channel callbacks\n",
            pENATSC3DASHFeeder->strDebugName);
    }

    free(pENATSC3DASHFeeder);

    return rval;
}

ENStatus
ENATSC3DASHFeederSetVerbosity(ENATSC3DASHFeederHandle_t hENATSC3DASHFeeder,
    ENVerbosity verbosity)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = 
        (ENATSC3DASHFeeder *)hENATSC3DASHFeeder;
    ENStatus rval = ENStatus_Success;

    pENATSC3DASHFeeder->verbosity = verbosity;

    return rval;
}

ENStatus
ENATSC3DASHFeederRegisterCallbacks(
    ENATSC3DASHFeederHandle_t hENATSC3DASHFeeder,
    ENATSC3DASHFeederMediaFileBeginCallback MediaFileBeginCallback,
    ENATSC3DASHFeederMediaFileEndCallback MediaFileEndCallback,
    ENATSC3DASHFeederMediaFileAbortCallback MediaFileAbortCallback,
    ENATSC3DASHFeederMediaFileDataCallback MediaFileDataCallback,
    ENATSC3DASHFeederLCTChannelRemovedCallback LCTChannelRemovedCallback,
    void *pCallbackUserData)
{
    ENATSC3DASHFeeder *pENATSC3DASHFeeder = 
        (ENATSC3DASHFeeder *)hENATSC3DASHFeeder;
    ENStatus rval = ENStatus_Success;

    pENATSC3DASHFeeder->MediaFileBeginCallback = MediaFileBeginCallback;
    pENATSC3DASHFeeder->MediaFileEndCallback = MediaFileEndCallback;
    pENATSC3DASHFeeder->MediaFileAbortCallback = MediaFileAbortCallback;
    pENATSC3DASHFeeder->MediaFileDataCallback = MediaFileDataCallback;
    pENATSC3DASHFeeder->LCTChannelRemovedCallback = LCTChannelRemovedCallback;
    pENATSC3DASHFeeder->pCallbackUserData = pCallbackUserData;

    return rval;
}
