/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include "libenatsc3ipv4.h"
#include "internetchecksum.h"

static char *ensys = "ipv4toudp";
static char *ensubsys = "main";

#define IPV4_HEADER_LENGTH 20
#define TRANSPORT_PACKET_MAX_SIZE 65536

#define ASSEMBLER_TIMEOUT_NS NANOSECONDS_PER_SECOND

typedef struct _ENATSC3IPV4Handle_t
{
    ENVerbosity verbosity;

    struct _ENATSC3IPV4TransportPacketAssembler
    {
        // The IP packet identifier for this transport packet. It is set to
        // -1 if no packet is being assembled on this assembler.
        int32_t identification;

        // The packet number of the first packet comprising the network 
        // layer packet being assembled.
        uint64_t packetNumber;

        // The plpId the first packet arrived on
        int32_t plpId;

        // The IPv4 address and protocol of the packet being assembled.
        uint32_t destinationAddress;
        uint32_t sourceAddress;
        uint16_t protocol;

        // Variables used in assembly of IPv4 fragments into transport layer
        // packets.
        uint8_t vTransportPacket[TRANSPORT_PACKET_MAX_SIZE];
        uint32_t nTransportPacketBytesUsed;
        uint32_t nextFragmentOffset;

        struct timespec packetTimeUTC;
        struct timespec packetTimeMonotonic;
    } vAssemblers[32];

    ENATSC3IPV4TransportPacketCallbackFunction TransportPacketCallback;
    void *pCallbackUserData;

    int ifd;
    int ofd;
} ENATSC3IPV4, *ENATSC3IPV4Handle_t;

static void
ReturnAssembler(ENATSC3IPV4 *pATSC3IPV4, 
    struct _ENATSC3IPV4TransportPacketAssembler *pAssembler)
{
    pAssembler->identification = -1;
}

static ENStatus
FindAssembler(ENATSC3IPV4 *pATSC3IPV4, uint32_t destinationAddress, 
    uint32_t sourceAddress, uint32_t identification, uint32_t protocol,
    struct _ENATSC3IPV4TransportPacketAssembler **ppAssembler)
{
    int32_t i;
    struct _ENATSC3IPV4TransportPacketAssembler *pAssembler = NULL;
    struct _ENATSC3IPV4TransportPacketAssembler *pIter;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pIter = pATSC3IPV4->vAssemblers; 
         i < numof(pATSC3IPV4->vAssemblers); 
         i++, pIter++)
    {
         if (pIter->identification == identification &&
             pIter->sourceAddress == sourceAddress &&
             pIter->destinationAddress == destinationAddress &&
             pIter->protocol == protocol)
         {
             // There is an existing assembler
             pAssembler = pIter;
             clock_gettime(CLOCK_MONOTONIC, &pAssembler->packetTimeMonotonic);
             break;
         }
    }

    if (!pAssembler)
    {
        rval = ENStatus_NotFound;
        goto out;
    }

    *ppAssembler = pAssembler;

out:
    return rval;
}

static ENStatus
AllocateAssembler(ENATSC3IPV4 *pATSC3IPV4, uint32_t destinationAddress, 
    uint32_t sourceAddress, uint32_t identification, uint32_t protocol,
    struct _ENATSC3IPV4TransportPacketAssembler **ppAssembler)
{
    int32_t i;
    struct _ENATSC3IPV4TransportPacketAssembler *pAssembler = NULL;
    struct _ENATSC3IPV4TransportPacketAssembler *pIter;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pIter = pATSC3IPV4->vAssemblers; 
         i < numof(pATSC3IPV4->vAssemblers); 
         i++, pIter++)
    {
         if (pIter->identification == identification &&
             pIter->sourceAddress == sourceAddress &&
             pIter->destinationAddress == destinationAddress &&
             pIter->protocol == protocol)
         {
             // There is an existing assembler
             LOGPRINT("Pre-existing assembler, reusing\n");
             pAssembler = pIter;
             break;
         }
         else if (pIter->identification == -1 && !pAssembler)
         {
             pAssembler = pIter;
         }
    }

    if (!pAssembler)
    {
        struct timespec mononow;
        clock_gettime(CLOCK_MONOTONIC, &mononow);

        for (i = 0, pIter = pATSC3IPV4->vAssemblers;
             i < numof(pATSC3IPV4->vAssemblers);
             i++, pIter++)
        {
            if ((int64_t)(mononow.tv_sec - pIter->packetTimeMonotonic.tv_sec) *
                NANOSECONDS_PER_SECOND + 
                (mononow.tv_nsec - pIter->packetTimeMonotonic.tv_nsec) > 
                ASSEMBLER_TIMEOUT_NS)
            {
                if (pATSC3IPV4->verbosity > ENVerbosity_Terse)
                {
                    LOGPRINT(
                        "Reclaiming timed out assembler %.8x/%.8x/%.4x/%d\n",
                        pIter->destinationAddress, pIter->sourceAddress, 
                        pIter->identification, pIter->protocol);
                }
                pAssembler = pIter;
                break;
            }
        }
    }

    if (!pAssembler)
    {
        ERRORPRINT("No free assemblers\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pAssembler->identification = identification;
    pAssembler->sourceAddress = sourceAddress;
    pAssembler->destinationAddress = destinationAddress;
    pAssembler->protocol = protocol;
    pAssembler->nextFragmentOffset = 0;
    pAssembler->nTransportPacketBytesUsed = 0;
    clock_gettime(CLOCK_MONOTONIC, &pAssembler->packetTimeMonotonic);

    *ppAssembler = pAssembler;

out:
    return rval;
}

#define IPV4Flags_DontFragment 0x2
#define IPV4Flags_MoreFragments 0x4

ENStatus
ENATSC3IPV4Open(ENATSC3IPV4Handle_t *phATSC3IPV4)
{
    ENATSC3IPV4 *pATSC3IPV4;
    struct _ENATSC3IPV4TransportPacketAssembler *pIter;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pATSC3IPV4 = (ENATSC3IPV4 *)calloc(sizeof(ENATSC3IPV4), 1);
    if (!pATSC3IPV4)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout; 
    }

    pATSC3IPV4->verbosity = ENVerbosity_Terse;

    for (i = 0, pIter = pATSC3IPV4->vAssemblers; 
         i < numof(pATSC3IPV4->vAssemblers); 
         i++, pIter++)
    {
        pIter->identification = -1;
    }

    *phATSC3IPV4 = pATSC3IPV4;

out:
    return rval;

errout:
    if (pATSC3IPV4)
    {
        free(pATSC3IPV4);
    }

    goto out;
}
    
ENStatus
ENATSC3IPV4Close(ENATSC3IPV4Handle_t hATSC3IPV4)
{
    ENATSC3IPV4 *pATSC3IPV4 = (ENATSC3IPV4 *)hATSC3IPV4;
    ENStatus rval = ENStatus_Success;

    free(pATSC3IPV4);

    return rval;
}

void
ENATSC3IPV4SetVerbosity(ENATSC3IPV4Handle_t hATSC3IPV4, 
    ENVerbosity verbosity)
{
    ENATSC3IPV4 *pATSC3IPV4 = (ENATSC3IPV4 *)hATSC3IPV4;

    pATSC3IPV4->verbosity = verbosity;

    return;
}

ENStatus
ENATSC3IPV4RegisterCallbacks(ENATSC3IPV4Handle_t hATSC3IPV4,
    ENATSC3IPV4TransportPacketCallbackFunction TransportPacketCallback, 
    void *pUserData)
{
    ENATSC3IPV4 *pATSC3IPV4 = (ENATSC3IPV4 *)hATSC3IPV4;
    ENStatus rval = ENStatus_Success;

    pATSC3IPV4->TransportPacketCallback = TransportPacketCallback;
    pATSC3IPV4->pCallbackUserData = pUserData;

    return rval;
}

ENStatus
ENATSC3IPV4ProcessIPV4Packet(ENATSC3IPV4Handle_t hATSC3IPV4, 
    int32_t plpId, struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const uint8_t *pPacket, int32_t packetLength)
{
    ENATSC3IPV4 *pATSC3IPV4 = (ENATSC3IPV4 *)hATSC3IPV4;
    struct _ENATSC3IPV4TransportPacketAssembler *pAssembler;
    int32_t version;
    int32_t internetHeaderLength;
    int32_t dscp;
    int32_t ecn;
    int32_t totalLength;
    int32_t identification;
    int32_t flags;
    int32_t fragmentOffset;
    int32_t ttl;
    int32_t protocol;
    int32_t headerChecksum;
    uint32_t sourceAddress;
    uint32_t destinationAddress;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    ENStatus rval = ENStatus_Success;

    if (packetLength < IPV4_HEADER_LENGTH)
    {
        if (pATSC3IPV4->verbosity >= ENVerbosity_Terse)
        {
            DEBUGPRINT("Short IPv4packet (%d < %d)\n", packetLength, 
                IPV4_HEADER_LENGTH);
        }
        goto out;
    }

    version = pPacket[0] >> 4;
    internetHeaderLength = pPacket[0] & 0xf;
    dscp = pPacket[1] >> 2;
    ecn = pPacket[1] & 0x3;
    totalLength = pPacket[2] << 8 | pPacket[3];
    identification = pPacket[4] << 8 | pPacket[5];
    flags = pPacket[6] >> 5;
    fragmentOffset = ((pPacket[6] & 0x1f) << 11) | (pPacket[7] << 3);
    ttl = pPacket[8];
    protocol = pPacket[9];
    headerChecksum = (pPacket[10] << 8) | pPacket[11];
    sourceAddress = ntohl(*(uint32_t *)(pPacket + 12));
    destinationAddress = ntohl(*(uint32_t *)(pPacket + 16));

    if (packetLength != totalLength)
    {
        if (pATSC3IPV4->verbosity >= ENVerbosity_Terse)
        {
            ERRORPRINT("IPv4 packet length mismatch %d != %d\n",
                totalLength, packetLength);
        }
        rval = ENStatus_InvalidData;
        goto out;
    }

    // Verify the header checksum
    ENInternetChecksumInitialize(&iccontext);
    ENInternetChecksumCalc(pPacket, internetHeaderLength * 4, &iccontext);
    sum = ENInternetChecksumFinalize(&iccontext);
    if (sum)
    {
        if (pATSC3IPV4->verbosity >= ENVerbosity_Terse)
        {
            ERRORPRINT("IPv4 header CHECKSUM ERROR: %x\n", sum);
        }
        rval = ENStatus_InvalidData;
        goto out;
    }

    if (fragmentOffset == 0)
    {
        if (AllocateAssembler(pATSC3IPV4, destinationAddress, sourceAddress, 
            identification, protocol, &pAssembler))
        {
            ERRORPRINT("Couldn't allocate assembler for %.8x/%.8x/%.4x/%d\n",
                destinationAddress, sourceAddress, identification, 
                protocol);
            rval = ENStatus_Dropped;
            goto out;
        }

        pAssembler->packetTimeUTC = *pPacketTimeUTC;
        pAssembler->packetNumber = packetNumber;
        pAssembler->plpId = plpId;
    }
    else
    {
        if (FindAssembler(pATSC3IPV4, destinationAddress, sourceAddress, 
            identification, protocol, &pAssembler))
        {
            if (pATSC3IPV4->verbosity >= ENVerbosity_Terse)
            {
                ERRORPRINT("No assembler for %.8x/%.8x/%.4x/%d\n",
                    destinationAddress, sourceAddress, identification, 
                    protocol);
            }
            rval = ENStatus_Dropped;
            goto out;
        }
    }

    // This IPv4 packet matches the transport packet we are currently 
    // assembling.
    if (pAssembler->nextFragmentOffset != fragmentOffset)
    {
        // The fragments are not being delivered in order, or fragments have 
        // been lost. We do not currently handle out-of-order fragments (they 
        // should not occur in broadcast ATSC 3.0 received at the ALP level, 
        // but can occur for broadband-delivered content or if the broadcast-
        // delivered ATSC 3.0 transits a LAN.)
        ERRORPRINT("Out-of-order fragments not supported (%d != %d) "
            " - Restarting assembly\n", 
            fragmentOffset, pAssembler->nextFragmentOffset);
        ReturnAssembler(pATSC3IPV4, pAssembler);
        rval = ENStatus_InvalidData;
        goto out;
    }

    if (packetLength < internetHeaderLength * 4)
    {
        if (pATSC3IPV4->verbosity >= ENVerbosity_Terse)
        {
            ERRORPRINT("Short packet (%d < %d)\n", packetLength, 
                internetHeaderLength * 4);
        }
        rval = ENStatus_InvalidData;
        goto out;
    }

    if (pAssembler->nTransportPacketBytesUsed + packetLength - 
        internetHeaderLength * 4 > sizeof(pAssembler->vTransportPacket))
    {
        ERRORPRINT("Transport datagram too large - restarting assembly\n");
        ReturnAssembler(pATSC3IPV4, pAssembler);
        rval = ENStatus_InvalidData;
        goto out;
    }

    memcpy(pAssembler->vTransportPacket + pAssembler->nTransportPacketBytesUsed,
        pPacket + internetHeaderLength * 4, 
        packetLength - internetHeaderLength * 4);
    pAssembler->nTransportPacketBytesUsed += 
        packetLength - internetHeaderLength * 4;
    pAssembler->nextFragmentOffset += 
        (packetLength - internetHeaderLength * 4) / 8;

    if (pATSC3IPV4->verbosity >= ENVerbosity_PacketTrace)
    {
        LOGPRINT("IPv4:\n");
        LOGPRINT("  version: %d\n", version);
        LOGPRINT("  IHL: %d (%d bytes)\n", internetHeaderLength, 
            internetHeaderLength * 4);
        LOGPRINT("  dscp: %d\n", dscp);
        LOGPRINT("  ecn: %d\n", ecn);
        LOGPRINT("  totalLength: %d\n", totalLength);
        LOGPRINT("  identification: %d\n", identification);
        LOGPRINT("  flags: %d\n", flags);
        LOGPRINT("  fragmentOffset: %d\n", fragmentOffset);
        LOGPRINT("  ttl: %d\n", ttl);
        LOGPRINT("  protocol: %d\n", protocol);
        LOGPRINT("  headerChecksum:%x \n", headerChecksum);
        LOGPRINT("  sourceAddress: %d.%d.%d.%d\n", 
            (sourceAddress >> 24) & 0xff,
            (sourceAddress >> 16) & 0xff,
            (sourceAddress >> 8) & 0xff,
            (sourceAddress & 0xff) & 0xff);
        LOGPRINT("  destinationAddress: %d.%d.%d.%d\n", 
            (destinationAddress >> 24) & 0xff,
            (destinationAddress >> 16) & 0xff,
            (destinationAddress >> 8) & 0xff,
            (destinationAddress) & 0xff);
    }

    // See if we've received the entire transport packet
    if (!(flags & IPV4Flags_MoreFragments))
    {
        if (pATSC3IPV4->TransportPacketCallback)
        {
            pATSC3IPV4->TransportPacketCallback(
                pAssembler->plpId,
                &pAssembler->packetTimeUTC,
                pAssembler->packetNumber,
                pAssembler->sourceAddress,
                pAssembler->destinationAddress,
                pAssembler->protocol,
                pAssembler->vTransportPacket, 
                pAssembler->nTransportPacketBytesUsed, 
                pATSC3IPV4->pCallbackUserData);
        }

        ReturnAssembler(pATSC3IPV4, pAssembler);
    }

out:
    return rval;
}
