/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_slt.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "slt";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static ENStatus
SLTInetUrlNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLTInetUrlNode *pSLTInetUrlNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_InetUrl);

    pSLTInetUrlNode = (ATSC3LLSSLTInetUrlNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSLTInetUrlNodeAttribute_urlType:
            if (EnXMLParser_antol((char *)pValue, sValue, 
                &pSLTInetUrlNode->urlType))
            {
                ERRORPRINT("Failed to parse urlType %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*surlType: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("    Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

static ENStatus
SLTInetUrlNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue, 
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData, 
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_SLTNode *pSLTNode = (ATSC3LLSSLT_SLTNode *)pUserData;
    ATSC3LLSSLTInetUrlNode *pSLTInetUrlNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_InetUrl);

    pSLTInetUrlNode = (ATSC3LLSSLTInetUrlNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(
        &pSLTNode->dataBuffer, (char *)pValue, sValue, 
        &pSLTInetUrlNode->urlDataOffset))
    {
        ERRORPRINT(
            "Failed to store url in string buffer\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*stext: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vSLTInetUrlAttributes[] = 
{
    {
        .id = ATSC3LLSSLTInetUrlNodeAttribute_urlType,
        .sName = sizeof("urlType") - 1, 
        .pName = "urlType",
        .handler = SLTInetUrlNodeAttributeHandler,
    },
};
EnXMLParserChildNode vSLTInetUrlChildNodes[] = 
{
};

EnXMLParserNode sltInetUrlNodeData = 
{
    .pName = "sltInetUrl",
    .pAttributes = vSLTInetUrlAttributes,
    .nAttributes = numof(vSLTInetUrlAttributes),
    .pChildNodes = vSLTInetUrlChildNodes,
    .nChildNodes = numof(vSLTInetUrlChildNodes),
    .textHandler = SLTInetUrlNodeTextHandler,
};

static ENStatus
ProcessInetUrlNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLTInetUrlNode *pInetUrlNode;
    ENStatus rval = ENStatus_Success;

    switch (pParentNode->nodeType)
    {
        case ATSC3LLSSLTNodeType_Service:
            {
                ATSC3LLSSLT_ServiceNode *pServiceNode;

                pServiceNode = (ATSC3LLSSLT_ServiceNode *)pParentNode;
                pInetUrlNode = &pServiceNode->svcInetUrl;
            }
            break; 
        case ATSC3LLSSLTNodeType_SLT:
            {
                ATSC3LLSSLT_SLTNode *pSLTNode;

                pSLTNode = (ATSC3LLSSLT_SLTNode *)pParentNode;
                pInetUrlNode = &pSLTNode->sltInetUrl;
            }
            break;
        default:
            ERRORPRINT("SLInetURl on unhandled parent %d\n", 
                pParentNode->nodeType);
            rval = ENStatus_Failure;
            goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*ssltInetUrl\n", xmlTextReaderDepth(reader) * 2, "");
    }
    pInetUrlNode->base.nodeType = ATSC3LLSSLTNodeType_InetUrl;

    rval = EnXMLParserProcessNode(&sltInetUrlNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)pInetUrlNode);

out:
    return rval;
}

static ENStatus
BroadcastSvcSignalingNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_BroadcastSvcSignalingNode *pBroadcastSvcSignalingNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_BroadcastSvcSignaling);

    pBroadcastSvcSignalingNode = 
        (ATSC3LLSSLT_BroadcastSvcSignalingNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pBroadcastSvcSignalingNode->slsProtocol))
            {
                ERRORPRINT("Failed to parse slsProtocol %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sslsProtocol: %ld\n", depth * 2, "", 
                    atol((char *)pValue));
            }
            break;
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMajorProtocolVersion:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pBroadcastSvcSignalingNode->slsMajorProtocolVersion))
            {
                ERRORPRINT("Failed to parse slsMajorProtocolVersion %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sslsMajorProtocolVersion: %ld\n", depth * 2, "",  
                    atol((char *)pValue));
            }
            break;
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMinorProtocolVersion:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                    &pBroadcastSvcSignalingNode->slsMinorProtocolVersion))
            {
                ERRORPRINT("Failed to parse slsMinorProtocolVersion %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sslsMinorProtocolVersion: %ld\n", depth * 2, "", 
                    atol((char *)pValue));
            }
            break;
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress:
            if (sValue >= 
                sizeof(pBroadcastSvcSignalingNode->slsDestinationIpAddress))
            {
                ERRORPRINT("IP address too long: %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            strncpy(pBroadcastSvcSignalingNode->slsDestinationIpAddress, 
                (char *)pValue, sValue);
            pBroadcastSvcSignalingNode->slsDestinationIpAddress[sValue] = 0;
            if (fVerbose)
            {
                LOGPRINT("%*sslsDestinationIpAddress: %.*s\n", depth * 2, "", 
                    sValue, pValue);
            }
            break;
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pBroadcastSvcSignalingNode->slsDestinationUdpPort))
            {
                ERRORPRINT("Failed to parse slsDestinationUdpPort %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sslsDestinationUdpPort: %ld\n", depth * 2, "", 
                    atol((char *)pValue));
            }
            break;
        case ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsSourceIpAddress:
            if (sValue >= 
                sizeof(pBroadcastSvcSignalingNode->slsSourceIpAddress))
            {
                ERRORPRINT("IP address too long: %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            strncpy(pBroadcastSvcSignalingNode->slsSourceIpAddress, 
                (char *)pValue, sValue);
            pBroadcastSvcSignalingNode->slsSourceIpAddress[sValue] = 0;
            if (fVerbose)
            {
                LOGPRINT("%*sslsSourceIpAddress: %.*s\n", depth * 2, "", 
                    sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                    pAttribute->id, pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vBroadcastSvcSignalingAttributes[] = 
{
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol,
        .sName = sizeof("slsProtocol") - 1, 
        .pName = "slsProtocol",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMajorProtocolVersion,
        .sName = sizeof("slsMajorProtocolVersion") - 1, 
        .pName = "slsMajorProtocolVersion",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMinorProtocolVersion,
        .sName = sizeof("slsMinorProtocolVersion") - 1, 
        .pName = "slsMinorProtocolVersion",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress,
        .sName = sizeof("slsDestinationIpAddress") - 1, 
        .pName = "slsDestinationIpAddress",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort,
        .sName = sizeof("slsDestinationUdpPort") - 1, 
        .pName = "slsDestinationUdpPort",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsSourceIpAddress,
        .sName = sizeof("slsSourceIpAddress") - 1, 
        .pName = "slsSourceIpAddress",
        .handler = BroadcastSvcSignalingNodeAttributeHandler,
    },
};

EnXMLParserChildNode vBroadcastSvcSignalingChildNodes[] = 
{
};

EnXMLParserNode broadcastSvcSignalingNodeData = 
{
    .pName = "BroadcastSvcSignaling",
    .pAttributes = vBroadcastSvcSignalingAttributes,
    .nAttributes = numof(vBroadcastSvcSignalingAttributes),
    .pChildNodes = vBroadcastSvcSignalingChildNodes,
    .nChildNodes = numof(vBroadcastSvcSignalingChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessBroadcastSvcSignalingNode(EnXMLParserChildNode *pChildNode,
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_ServiceNode *pServiceNode;
    ATSC3LLSSLT_BroadcastSvcSignalingNode *pBroadcastSvcSignalingNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_Service);

    pServiceNode = (ATSC3LLSSLT_ServiceNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    pBroadcastSvcSignalingNode = &pServiceNode->broadcastSvcSignaling;
    pBroadcastSvcSignalingNode->base.nodeType = 
        ATSC3LLSSLTNodeType_BroadcastSvcSignaling;

    if (fVerbose)
    {
        LOGPRINT("%*sBroadcastSvcSignaling\n", depth * 2, "");
    }

    rval = EnXMLParserProcessNode(&broadcastSvcSignalingNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)pBroadcastSvcSignalingNode);

    return rval;
}

static ENStatus
OtherBsidNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_OtherBsidNode *pOtherBsidNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_OtherBsid);

    pOtherBsidNode = (ATSC3LLSSLT_OtherBsidNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSLTOtherBsidNodeAttribute_type:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pOtherBsidNode->type))
            {
                ERRORPRINT("Failed to parse type %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*stype: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

static ENStatus
OtherBsidNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue, 
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData, 
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_OtherBsidNode *pOtherBsidNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_OtherBsid);

    pOtherBsidNode = (ATSC3LLSSLT_OtherBsidNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParser_antoul((char *)pValue, sValue, &pOtherBsidNode->otherBsid))
    {
        ERRORPRINT("Failed to parse otherBsid %.*s\n", sValue, pValue);
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*stext: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vOtherBsidAttributes[] = 
{
    {
        .id = ATSC3LLSSLTOtherBsidNodeAttribute_type,
        .sName = sizeof("type") - 1, 
        .pName = "type",
        .handler = OtherBsidNodeAttributeHandler,

    },
};

EnXMLParserChildNode vOtherBsidChildNodes[] = 
{
};

EnXMLParserNode otherBsidNodeData = 
{
    .pName = "OtherBsid",
    .pAttributes = vOtherBsidAttributes,
    .nAttributes = numof(vOtherBsidAttributes),
    .pChildNodes = vOtherBsidChildNodes,
    .nChildNodes = numof(vOtherBsidChildNodes),
    .textHandler = OtherBsidNodeTextHandler,
};

static ENStatus
ProcessOtherBsidNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_SLTNode *pSLTNode = (ATSC3LLSSLT_SLTNode *)pUserData;
    ATSC3LLSSLT_ServiceNode *pServiceNode;
    ATSC3LLSSLT_OtherBsidNode otherBsidNode;
    int32_t otherBsidNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_Service);
    pServiceNode = (ATSC3LLSSLT_ServiceNode *)pParentNode;

    memset(&otherBsidNode, 0, sizeof(otherBsidNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the rating node's ratingValueAbbrevs data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSLTNode->dataBuffer, 
        sizeof(otherBsidNode), &otherBsidNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for otherBsid node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pServiceNode->otherBsids, 
        otherBsidNodeOffset))
    {
        ERRORPRINT("Failed to store otherBsid node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*OtherBsid\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    otherBsidNode.base.nodeType = ATSC3LLSSLTNodeType_OtherBsid;
    rval = EnXMLParserProcessNode(&otherBsidNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&otherBsidNode);

    memcpy(pSLTNode->dataBuffer.pBuffer + otherBsidNodeOffset, &otherBsidNode,
        sizeof(otherBsidNode));

out:
    return rval;
}

static ENStatus
ServiceNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_SLTNode *pSLTNode = (ATSC3LLSSLT_SLTNode *)pUserData;
    ATSC3LLSSLT_ServiceNode *pServiceNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_Service);

    pServiceNode = (ATSC3LLSSLT_ServiceNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSLTServiceNodeAttribute_serviceId:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceNode->serviceId))
            {
                ERRORPRINT("Failed to parse serviceId %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceId: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_sltSvcSeqNum:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceNode->sltSvcSeqNum))
            {
                ERRORPRINT("Failed to parse sltSvcSeqNum %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*ssltSvcSeqNum: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_globalServiceID:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSLTNode->dataBuffer, (char *)pValue, sValue, 
                &pServiceNode->globalServiceIDDataOffset))
            {
                ERRORPRINT(
                    "Failed to store globalServiceID in string buffer\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sglobalServiceID: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_protected:
            if (EnXMLParserStringToBool((char *)pValue, sValue,
                &pServiceNode->protected))
            {
                ERRORPRINT("Failed to eval protected\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sprotected: %.*s\n", depth * 2, "", sValue, 
                    pValue);  
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_majorChannelNo:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceNode->majorChannelNo))
            {
                ERRORPRINT("Failed to parse majorChannelNo %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*smajorChannelNo: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_minorChannelNo:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceNode->minorChannelNo))
            {
                ERRORPRINT("Failed to parse minorChannelNo %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sminorChannelNo: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_serviceCategory:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pServiceNode->serviceCategory))
            {
                ERRORPRINT("Failed to parse serviceCategory %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceCategory: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_shortServiceName:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSLTNode->dataBuffer, (char *)pValue, sValue, 
                &pServiceNode->shortServiceNameDataOffset))
            {
                ERRORPRINT(
                    "Failed to store shortServiceName in string buffer\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sshortServiceName: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_hidden:
            if (EnXMLParserStringToBool((char *)pValue, sValue,
                &pServiceNode->hidden))
            {
                ERRORPRINT("Failed to eval hidden\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*shidden: %.*s\n", depth * 2, "", sValue, pValue); 
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_essential:
            if (EnXMLParserStringToBool((char *)pValue, sValue,
                &pServiceNode->essential))
            {
                ERRORPRINT("Failed to eval essential\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sessential: %.*s\n", depth * 2, "", sValue, 
                    pValue); 
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_broadbandAccessRequired:
            if (EnXMLParserStringToBool((char *)pValue, sValue,
                &pServiceNode->broadbandAccessRequired))
            {
                ERRORPRINT("Failed to eval broadbandAccessRequired\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sbroadbandAccessRequired: %.*s\n", depth * 2, "",  
                    sValue, pValue);
            }
            break;
        case ATSC3LLSSLTServiceNodeAttribute_svcCapabilities:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSLTNode->dataBuffer, (char *)pValue, sValue, 
                &pServiceNode->svcCapabilitiesDataOffset))
            {
                ERRORPRINT(
                    "Failed to store svcCapabilities in string buffer\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*ssvcCapabilities: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("    Unhandled attribute %d (%s)\n", 
                    pAttribute->id, pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vServiceAttributes[] = 
{
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_serviceId,
        .sName = sizeof("serviceId") - 1, 
        .pName = "serviceId",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_sltSvcSeqNum,
        .sName = sizeof("sltSvcSeqNum") - 1, 
        .pName = "sltSvcSeqNum",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_globalServiceID,
        .sName = sizeof("globalServiceID") - 1, 
        .pName = "globalServiceID",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_protected,
        .sName = sizeof("protected") - 1, 
        .pName = "protected",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_majorChannelNo,
        .sName = sizeof("majorChannelNo") - 1, 
        .pName = "majorChannelNo",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_minorChannelNo,
        .sName = sizeof("minorChannelNo") - 1, 
        .pName = "minorChannelNo",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_serviceCategory,
        .sName = sizeof("serviceCategory") - 1, 
        .pName = "serviceCategory",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_shortServiceName,
        .sName = sizeof("shortServiceName") - 1, 
        .pName = "shortServiceName",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_hidden,
        .sName = sizeof("hidden") - 1, 
        .pName = "hidden",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_essential,
        .sName = sizeof("essential") - 1, 
        .pName = "essential",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_broadbandAccessRequired,
        .sName = sizeof("broadbandAccessRequired") - 1, 
        .pName = "broadbandAccessRequired",
        .handler = ServiceNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTServiceNodeAttribute_svcCapabilities,
        .sName = sizeof("svcCapabilities") - 1, 
        .pName = "svcCapabilities",
        .handler = ServiceNodeAttributeHandler,
    },
};
EnXMLParserChildNode vServiceChildNodes[] = 
{
    {
        .id = ATSC3LLSSLTServiceNodeChildNode_BroadcastSvcSignaling,
        .sName = sizeof("BroadcastSvcSignaling") - 1,
        .pName = "BroadcastSvcSignaling",
        .handler = ProcessBroadcastSvcSignalingNode,
    },
    {
        .id = ATSC3LLSSLTServiceNodeChildNode_svcInetUrl,
        .sName = sizeof("svcInetUrl") - 1,
        .pName = "svcInetUrl",
        .handler = ProcessInetUrlNode,
    },
    {
        .id = ATSC3LLSSLTServiceNodeChildNode_otherBsid,
        .sName = sizeof("otherBsid") - 1,
        .pName = "otherBsid",
        .handler = ProcessOtherBsidNode,
    }
};
EnXMLParserNode serviceNodeData = 
{
    .pName = "Service",
    .pAttributes = vServiceAttributes,
    .nAttributes = numof(vServiceAttributes),
    .pChildNodes = vServiceChildNodes,
    .nChildNodes = numof(vServiceChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessServiceNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_SLTNode *pSLTNode;
    ATSC3LLSSLT_ServiceNode serviceNode;
    int32_t serviceNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_SLT);
    pSLTNode = (ATSC3LLSSLT_SLTNode *)pParentNode;

    memset(&serviceNode, 0, sizeof(serviceNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the rating node's ratingValueAbbrevs data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pSLTNode->dataBuffer, 
        sizeof(serviceNode), &serviceNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for service node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pSLTNode->services, 
        serviceNodeOffset))
    {
        ERRORPRINT("Failed to store services node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&serviceNode.otherBsids,
        &pSLTNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate otherBsids array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*Services\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    serviceNode.base.nodeType = ATSC3LLSSLTNodeType_Service;
    rval = EnXMLParserProcessNode(&serviceNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&serviceNode);

    memcpy(pSLTNode->dataBuffer.pBuffer + serviceNodeOffset, &serviceNode,
        sizeof(serviceNode));

out:
    return rval;
}

static ENStatus
SLTNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSLT_SLTNode *pSLTNode = (ATSC3LLSSLT_SLTNode *)pUserData;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSLTNodeType_SLT);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSLTNodeAttribute_BSID:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pSLTNode->bsid))
            {
                ERRORPRINT("Failed to parse bsid %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sbsid: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSLTNodeAttribute_SLTCapabilities:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSLTNode->dataBuffer, (char *)pValue, sValue, 
                &pSLTNode->sltCapabilitiesDataOffset))
            {
                ERRORPRINT(
                    "Failed to store capabilities in string buffer\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*ssltCapabilities: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSLTNodeAttribute_XMLNS:
            if (fVerbose)
            {
                LOGPRINT("%*sxmlns: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", 
                    pAttribute->id, pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vSLTAttributes[] = 
{
    {
        .id = ATSC3LLSSLTNodeAttribute_BSID, 
        .sName = sizeof("bsid") - 1, 
        .pName = "bsid",
        .handler = SLTNodeAttributeHandler,

    },
    {
        .id = ATSC3LLSSLTNodeAttribute_SLTCapabilities, 
        .sName = sizeof("sltCapabilities") - 1, 
        .pName = "sltCapabilities",
        .handler = SLTNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSLTNodeAttribute_XMLNS, 
        .sName = sizeof("xmlns") - 1, 
        .pName = "xmlns",
        .handler = SLTNodeAttributeHandler,
    },
};
EnXMLParserChildNode vSLTChildNodes[] = 
{
    {
        .id = ATSC3LLSSLTNodeChildNode_SLTInetUrl,
        .sName = sizeof("sltInetUrl") - 1,
        .pName = "sltInetUrl",
        .handler = ProcessInetUrlNode,
    },
    {
        .id = ATSC3LLSSLTNodeChildNode_Service,
        .sName = sizeof("Service") - 1,
        .pName = "Service",
        .handler = ProcessServiceNode,
    }
};
EnXMLParserNode sltNodeData = 
{
    .pName = "SLT",
    .pAttributes = vSLTAttributes,
    .nAttributes = numof(vSLTAttributes),
    .pChildNodes = vSLTChildNodes,
    .nChildNodes = numof(vSLTChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessSLTNode(xmlTextReaderPtr reader, ATSC3LLSSLT_SLTNode *pSLTNode)
{
    ENStatus rval = ENStatus_Success;

    if (fVerbose)
    {
        LOGPRINT("SLT\n");
    }

    if (EnXMLParserDataOffsetArrayInitialize(&pSLTNode->services,
        &pSLTNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingRegionTables array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&sltNodeData, reader, pSLTNode,
        (EnXMLParserNodeBase *)pSLTNode);

out:
    return rval;
}

ENStatus
ATSC3LLSParseSLT(const char *pSLT, int32_t sSLT, 
    ATSC3LLSSLT_SLTNode *pParsedSLT)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedSLT, 0, sizeof(ATSC3LLSSLT_SLTNode));
    pParsedSLT->base.nodeType = ATSC3LLSSLTNodeType_SLT;

    if (EnXMLParserDataBufferInitialize(&pParsedSLT->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to allocate SLT data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pSLT, sSLT, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader (size %d {%.*s})\n", sSLT,
            sSLT, pSLT);
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "SLT"))
        {
            int32_t lret;

            lret = ProcessSLTNode(reader, pParsedSLT);
            if (lret)
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't SLT (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse SLT XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedSLT->dataBuffer))
        {
            ERRORPRINT("Failed to free SLT data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3LLSFreeParsedSLT(ATSC3LLSSLT_SLTNode *pParsedSLT)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedSLT->dataBuffer))
    {
        ERRORPRINT("Failed to free SLT data buffer\n");
    }

    return rval;
}
