/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of
 * the following permission added to Section 15 as permitted in Section 7(a):
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libendebug/libendebug.h>
#include <enatsc3-common/enatsc3_specdefs.h>

#include <enatsc3-common/enatsc3.h>
#include "libenatsc3lls.h"

static char *ensys = "lls";
static char *ensubsys = "main";

ENStatus
ENATSC3LLSProcessSignedMultiTable(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const ENATSC3LLSTableHeader *pLLSTableHeader, 
    const uint8_t *pPayload, int32_t payloadLength,
    ENATSC3LLSTableCallbackFunction LLSTableCallback, void *pUserData)
{
    uint8_t lls_payload_count;
    const uint8_t *pOPayload;
    int32_t oPayloadLength;
    int32_t signatureLength;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pOPayload = pPayload;
    oPayloadLength = payloadLength;

    // We are first going to verify that the signed multitable is large
    // enough to hold the tables and the signature. The app should
    // have called ENATSC3SSPVerifySignedMultitableSignature which
    // would have verified this, but just in case we do it again here.
    if (!payloadLength)
    {
        ERRORPRINT("SignedMultiTable too short\n");
        goto out;
    }
    lls_payload_count = pPayload[0];
    payloadLength--;
    pPayload++;

    for (i = 0; i < lls_payload_count; i++)
    {
        int32_t lls_payload_length;

        if (payloadLength < 4)
        {
            ERRORPRINT("SignedMultiTable: table %d header truncated\n", i);
            rval = ENStatus_Failure;
            goto out;
        }
        lls_payload_length = (pPayload[2] << 8) | pPayload[3];
        payloadLength -= 4;
        pPayload += 4;

        if (payloadLength < lls_payload_length)
        {
            ERRORPRINT("SignedMultiTable: payload %d trucated\n", i);
        }

        payloadLength -= lls_payload_length;
        pPayload += lls_payload_length;
    }

    if (payloadLength < 2)
    {
        ERRORPRINT("SignedMultiTable: failed to get signature length\n");
        rval = ENStatus_Failure;
        goto out;
    }

    signatureLength = (pPayload[0] << 8) | pPayload[1];

    if (payloadLength < 2 + signatureLength)
    {
        ERRORPRINT("SignedMultiTable: truncated signature\n");
        rval = ENStatus_Failure;
        goto out;
    }

    // The table length checks out, now process the tables.
    pPayload = pOPayload;
    payloadLength = oPayloadLength;

    lls_payload_count = pPayload[0];
    payloadLength--;
    pPayload++;

    for (i = 0; i < lls_payload_count; i++)
    {
        ENATSC3LLSTableHeader llsTableHeader;
        int32_t lls_payload_length;

        llsTableHeader = *pLLSTableHeader;
        llsTableHeader.lls_table_id = pPayload[0];
        llsTableHeader.lls_table_version = pPayload[1];
        lls_payload_length = (pPayload[2] << 8) | pPayload[3];

        payloadLength -= 4;
        pPayload += 4;

        LLSTableCallback(pPacketTimeUTC, packetNumber, &llsTableHeader, 
            pPayload, lls_payload_length, 1 /* fSigned */, pUserData);

        payloadLength -= lls_payload_length;
        pPayload += lls_payload_length;
    }

out:
    return rval;
}

