/*
 * Copyright © 2019-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3sls_held.h"

static char *ensys = "sls";
static char *ensubsys = "held";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
HTMLEntryPackageNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSHELDNode *pHELDNode = (ENATSC3SLSHELDNode *)pUserData;
    ENATSC3SLSHELDHTMLEntryPackageNode *pHTMLEntryPackageNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3SLSHELDNodeType_HTMLEntryPackage);

    pHTMLEntryPackageNode = (ENATSC3SLSHELDHTMLEntryPackageNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ENATSC3SLSHELDHTMLEntryPackage_appContextId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->appContextIdDataOffset))
            {
                ERRORPRINT("Failed to store appContextId\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sappContextId: %.*s\n", depth * 2,  "",
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_requiredCapabilities:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->requiredCapabilitiesDataOffset))
            {
                ERRORPRINT("Failed to store requiredCapabilities\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*srequiredCapabilities: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_appRendering:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pHTMLEntryPackageNode->appRendering))
            {
                ERRORPRINT("Failed to parse appRendering %.*s\n", sValue, 
                    pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sappRendering: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_clearAppContextCacheDate:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue, 
                &pHTMLEntryPackageNode->clearAppContextCacheDateSecondsUTC, 
                NULL))
            {
                ERRORPRINT("Failed to parse clearAppContextCacheDate %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sclearAppContextCacheDate: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPackageUrl:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->bcastEntryPackageUrlDataOffset))
            {
                ERRORPRINT("Failed to store bcastEntryPackageUrl\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sbcastEntryPackageUrl: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPageUrl:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->bcastEntryPageUrlDataOffset))
            {
                ERRORPRINT("Failed to store bcastEntryPageUrl\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sbcastEntryPageUrl: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_bbandEntryPageUrl:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->bbandEntryPageUrlDataOffset))
            {
                ERRORPRINT("Failed to store bbandEntryPageUrl\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sbbandEntryPageUrl: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_validFrom:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue, 
                &pHTMLEntryPackageNode->validFromSecondsUTC, NULL))
            {
                ERRORPRINT("Failed to parse validFrom %.*s\n", sValue, 
                    pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*svalidFrom: %.*s\n", depth * 2, "",
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_validUntil:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue, 
                &pHTMLEntryPackageNode->validUntilSecondsUTC, NULL))
            {
                ERRORPRINT("Failed to parse validUntil %.*s\n", sValue, 
                    pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*svalidUntil: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_coupledServices:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->coupledServicesDataOffset))
            {
                ERRORPRINT("Failed to store coupledServices\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*scoupledServices: %.*s\n", depth * 2, "",
                sValue, pValue);
            break;
        case ENATSC3SLSHELDHTMLEntryPackage_lctTSIRef:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pHELDNode->dataBuffer, (char *)pValue, sValue, 
                &pHTMLEntryPackageNode->lctTSIRefDataOffset))
            {
                ERRORPRINT("Failed to store lctTSIRef\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*slctTSIRef: %.*s\n", depth * 2, "", 
                sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vHTMLEntryPackageAttributes[] = 
{
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_appContextId,
        .sName = sizeof("appContextId") - 1,
        .pName = "appContextId",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_requiredCapabilities,
        .sName = sizeof("requiredCapabilities") - 1,
        .pName = "requiredCapabilities",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_appRendering,
        .sName = sizeof("appRendering") - 1,
        .pName = "appRendering",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_clearAppContextCacheDate,
        .sName = sizeof("clearAppContextCacheDate") - 1,
        .pName = "clearAppContextCacheDate",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPackageUrl,
        .sName = sizeof("bcastEntryPackageUrl") - 1,
        .pName = "bcastEntryPackageUrl",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPageUrl,
        .sName = sizeof("bcastEntryPageUrl") - 1,
        .pName = "bcastEntryPageUrl",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_bbandEntryPageUrl,
        .sName = sizeof("bbandEntryPageUrl") - 1,
        .pName = "bbandEntryPageUrl",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_validFrom,
        .sName = sizeof("validFrom") - 1,
        .pName = "validFrom",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_validUntil,
        .sName = sizeof("validUntil") - 1,
        .pName = "validUntil",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_coupledServices,
        .sName = sizeof("coupledServices") - 1,
        .pName = "coupledServices",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
    {
        .id = ENATSC3SLSHELDHTMLEntryPackage_lctTSIRef,
        .sName = sizeof("lctTSIRef") - 1,
        .pName = "lctTSIRef",
        .handler = HTMLEntryPackageNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vHTMLEntryPackageChildNodes[] =
{
};
static EnXMLParserNode htmlEntryPackageNodeData = 
{
    .pName = "HTMLEntryPackage",
    .pAttributes = vHTMLEntryPackageAttributes,
    .nAttributes = numof(vHTMLEntryPackageAttributes),
    .pChildNodes = vHTMLEntryPackageChildNodes,
    .nChildNodes = numof(vHTMLEntryPackageChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessHTMLEntryPackageNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3SLSHELDNode *pHELDNode = (ENATSC3SLSHELDNode *)pUserData;
    ENATSC3SLSHELDHTMLEntryPackageNode htmlEntryPackageNode;
    int32_t htmlEntryPackageNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3SLSHELDNodeType_HELD);
    memset(&htmlEntryPackageNode, 0, sizeof(htmlEntryPackageNode));

    // Allocate space for the node from the data buffer and store the offset 
    // in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pHELDNode->dataBuffer,
        sizeof(ENATSC3SLSHELDHTMLEntryPackageNode), &htmlEntryPackageNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for htmlEntryPackageNode node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pHELDNode->htmlEntryPackages, 
        htmlEntryPackageNodeOffset))
    {
        ERRORPRINT("Failed to store htmlEntryPackage node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    htmlEntryPackageNode.base.nodeType = ENATSC3SLSHELDNodeType_HTMLEntryPackage;

    VERBOSEPRINT("%*sHTMLEntryPackage\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&htmlEntryPackageNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&htmlEntryPackageNode);

    memcpy(pHELDNode->dataBuffer.pBuffer + htmlEntryPackageNodeOffset,
        &htmlEntryPackageNode, sizeof(htmlEntryPackageNode));

out:
    return rval;
}

static EnXMLParserAttribute vHELDAttributes[] = 
{
};
static EnXMLParserChildNode vHELDChildNodes[] =
{
    {
        .id = ENATSC3SLSHELDNodeType_HTMLEntryPackage,
        .sName = sizeof("HTMLEntryPackage") - 1,
        .pName = "HTMLEntryPackage",
        .handler = ProcessHTMLEntryPackageNode,
    },
};
static EnXMLParserNode heldNodeData = 
{
    .pName = "HELD",
    .pAttributes = vHELDAttributes,
    .nAttributes = numof(vHELDAttributes),
    .pChildNodes = vHELDChildNodes,
    .nChildNodes = numof(vHELDChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessHELDNode(xmlTextReaderPtr reader, ENATSC3SLSHELDNode *pHELDNode)
{
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*sHELD\n", xmlTextReaderDepth(reader) * 2, "");

    if (EnXMLParserDataOffsetArrayInitialize(&pHELDNode->htmlEntryPackages,
        &pHELDNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate htmlEntryPackages array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&heldNodeData, reader, pHELDNode,
        (EnXMLParserNodeBase *)pHELDNode);

out:
    return rval;
}

ENStatus
ENATSC3SLSHELDParse(const char *pXML, int32_t sXML, 
    ENATSC3SLSHELDNode *pParsedHELD)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedHELD, 0, sizeof(ENATSC3SLSHELDNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedHELD->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to initialize HELD data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "HELD"))
        {
            pParsedHELD->base.nodeType = ENATSC3SLSHELDNodeType_HELD;
            ProcessHELDNode(reader, pParsedHELD);
        }
        else
        {
            ERRORPRINT("Top-level object isn't HELD (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse HELD XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedHELD->dataBuffer))
        {
            ERRORPRINT("Failed to free HELD data buffer\n");
        }
    }

    goto out;
}

ENStatus
ENATSC3SLSHELDFree(ENATSC3SLSHELDNode *pParsedHELD)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedHELD->dataBuffer))
    {
        ERRORPRINT("Failed to free HELD data buffer\n");
    }

    return rval;
}

